"""Test unit for app status and counters."""

import pytest

import zigpy.state as app_state

COUNTER_NAMES = ["counter_1", "counter_2", "some random name"]


@pytest.fixture
def counters():
    """Counters fixture."""
    counters = app_state.CounterGroup("ezsp_counters")
    for name in COUNTER_NAMES:
        counters[name]
    return counters


def test_counter():
    """Test basic counter."""

    counter = app_state.Counter("mock_counter")
    assert counter.value == 0

    counter = app_state.Counter("mock_counter", 5)
    assert counter.value == 5
    assert counter.reset_count == 0

    counter.update(5)
    assert counter.value == 5
    assert counter.reset_count == 0

    counter.update(8)
    assert counter.value == 8
    assert counter.reset_count == 0

    counter.update(9)
    assert counter.value == 9
    assert counter.reset_count == 0

    counter.reset()
    assert counter.value == 9
    assert counter._raw_value == 0
    assert counter.reset_count == 1

    # new value after a counter was reset/clear
    counter.update(12)
    assert counter.value == 21
    assert counter.reset_count == 1

    counter.update(15)
    assert counter.value == 24
    assert counter.reset_count == 1

    # new counter value is less than previously reported.
    # assume counter was reset
    counter.update(14)
    assert counter.value == 24 + 14
    assert counter.reset_count == 2

    counter.reset_and_update(14)
    assert counter.value == 38 + 14
    assert counter.reset_count == 3


def test_counter_str():
    """Test counter str representation."""

    counter = app_state.Counter("some_counter", 8)
    assert str(counter) == "some_counter = 8"


def test_counters_init():
    """Test counters initialization."""

    counter_groups = app_state.CounterGroups()
    assert len(counter_groups) == 0
    counters = counter_groups["ezsp_counters"]
    assert len(counter_groups) == 1

    assert len(counters) == 0
    assert counters.name == "ezsp_counters"
    for name in COUNTER_NAMES:
        counters[name]
    assert len(counters) == 3

    cnt_1, cnt_2, cnt_3 = (counter for counter in counters.counters())
    assert cnt_1.name == "counter_1"
    assert cnt_2.name == "counter_2"
    assert cnt_3.name == "some random name"

    assert cnt_1.value == 0
    assert cnt_2.value == 0
    assert cnt_3.value == 0

    counters["some random name"].update(2)
    assert cnt_3.value == 2
    assert counters["some random name"].value == 2
    assert counters["some random name"] == 2
    assert counters["some random name"] == cnt_3
    assert int(cnt_3) == 2

    assert "counter_2" in counters
    assert [counter.name for counter in counters.counters()] == COUNTER_NAMES

    counters.reset()
    for counter in counters.counters():
        assert counter.reset_count == 1


def test_counters_str_and_repr(counters):
    """Test counters str and repr."""

    counters["counter_1"].update(22)
    counters["counter_2"].update(33)

    assert (
        str(counters)
        == "ezsp_counters: [counter_1 = 22, counter_2 = 33, some random name = 0]"
    )

    assert (
        repr(counters)
        == """CounterGroup('ezsp_counters', {Counter('counter_1', 22), """
        """Counter('counter_2', 33), Counter('some random name', 0)})"""
    )


def test_state():
    """Test state structure."""
    state = app_state.State()
    assert state
    assert state.counters == {}

    assert state.counters["new_collection"]["counter_2"] == 0
    assert state.counters["new_collection"]["counter_2"].reset_count == 0
    assert state.counters["new_collection"]["counter_3"].reset_count == 0
    state.counters["new_collection"]["counter_2"] = 2


def test_counters_reset(counters):
    """Test counter resetting."""

    counter = counters["counter_1"]

    assert counter.reset_count == 0
    counters["counter_1"].update(22)
    assert counter.value == 22
    assert counter.reset_count == 0

    counters.reset()
    assert counter.reset_count == 1
    counter.update(22)
    assert counter.value == 44
    assert counter.reset_count == 1


def test_counter_incr():
    """Test counter increment."""

    counter = app_state.Counter("counter_name", 42)
    assert counter == 42

    counter.increment()
    assert counter == 43

    counter.increment(5)
    assert counter == 48
    assert counter.value == 48

    with pytest.raises(AssertionError):
        counter.increment(-1)


def test_counter_nested_groups_increment():
    """Test nested counters."""

    counters = app_state.CounterGroup("device_counters")

    assert len(counters) == 0
    counters.increment("reply", "rx", "zdo", 0x8031)
    counters.increment("total", "rx", 3, 0x0006)
    counters.increment("total", "rx", 3, 0x0008)
    counters.increment("total", "rx", 3, 0x0300)

    tags = set(counters.tags())
    assert {"rx"} == tags

    tags = set(counters["rx"].tags())
    assert {"zdo", 3} == tags

    assert counters["rx"]["reply"] == 1
    assert counters["rx"]["zdo"]["reply"] == 1
    assert counters["rx"]["zdo"][0x8031]["reply"] == 1

    assert counters["rx"]["total"] == 3
    assert counters["rx"][3]["total"] == 3
    assert counters["rx"][3][0x0006]["total"] == 1
    assert counters["rx"][3][0x0008]["total"] == 1
    assert counters["rx"][3][0x0300]["total"] == 1


def test_counter_groups():
    """Test CounterGroups."""

    groups = app_state.CounterGroups()
    assert not list(groups)

    counter_group = groups["ezsp_counters"]

    new_groups = list(groups)
    assert new_groups == [counter_group]



from __future__ import annotations

import asyncio
import fcntl
import pathlib
from unittest.mock import AsyncMock, Mock, call, patch

import pytest

import zigpy.serial
from zigpy.typing import UNDEFINED, UndefinedType


# fmt: off
@pytest.mark.parametrize(("url", "flow_control", "xonxoff", "rtscts", "expected_kwargs"), [
    # `flow_control` on its own
    ("/dev/ttyUSB1", "hardware", UNDEFINED, UNDEFINED, {"xonxoff": False, "rtscts": True}),
    ("/dev/ttyUSB1", "software", UNDEFINED, UNDEFINED, {"xonxoff": True,  "rtscts": False}),
    ("/dev/ttyUSB1", None,       UNDEFINED, UNDEFINED, {"xonxoff": False, "rtscts": False}),

    # `flow_control` overrides `xonxoff` and `rtscts`
    ("/dev/ttyUSB1", "hardware", True,      False,     {"xonxoff": False, "rtscts": True}),
    ("/dev/ttyUSB1", "software", False,      True,     {"xonxoff": True,  "rtscts": False}),
    ("/dev/ttyUSB1", None,       True,      False,     {"xonxoff": False, "rtscts": False}),

    # `flow_control` defaults to undefined so `xonxoff` and `rtscts` are used
    ("/dev/ttyUSB1", UNDEFINED,  True,      False,     {"xonxoff": True,  "rtscts": False}),
    ("/dev/ttyUSB1", UNDEFINED,  False,      True,     {"xonxoff": False, "rtscts": True}),
    ("/dev/ttyUSB1", UNDEFINED,  True,       True,     {"xonxoff": True,  "rtscts": True}),

    # The defaults are used when `flow_control`, `xonxoff`, and `rtscts` are all undefined
    ("/dev/ttyUSB1", UNDEFINED,  UNDEFINED, UNDEFINED, {"xonxoff": False, "rtscts": False}),
])
# fmt: on
async def test_serial_normal(
    url: str,
    flow_control: str | UndefinedType,
    xonxoff: bool | UndefinedType,
    rtscts: bool | UndefinedType,
    expected_kwargs: dict[str, bool],
) -> None:
    loop = asyncio.get_running_loop()
    protocol_factory = Mock()

    kwargs = {"url": url}

    if flow_control is not UNDEFINED:
        kwargs["flow_control"] = flow_control

    if xonxoff is not UNDEFINED:
        kwargs["xonxoff"] = xonxoff

    if rtscts is not UNDEFINED:
        kwargs["rtscts"] = rtscts

    with patch(
        "zigpy.serial.serialx_create_serial_connection",
        AsyncMock(
            return_value=(AsyncMock(), AsyncMock())
        ),
    ) as mock_create_serial_connection:
        await zigpy.serial.create_serial_connection(loop, protocol_factory, **kwargs)

    mock_calls = mock_create_serial_connection.mock_calls
    assert len(mock_calls) == 1

    assert mock_calls[0].kwargs["url"] == "/dev/ttyUSB1"
    assert mock_calls[0].kwargs["baudrate"] == 115200

    for kwarg in expected_kwargs:
        assert mock_calls[0].kwargs[kwarg] == expected_kwargs[kwarg]


async def test_serial_error_remapping(tmp_path: pathlib.Path) -> None:
    loop = asyncio.get_running_loop()
    protocol_factory = Mock()

    # FileNotFoundError
    missing_port = tmp_path / "missing"
    assert not missing_port.exists()

    with pytest.raises(FileNotFoundError):
        await zigpy.serial.create_serial_connection(
            loop, protocol_factory, url=missing_port
        )

    # PermissionError
    denied_port = tmp_path / "denied"
    denied_port.touch()
    denied_port.chmod(0o000)

    with pytest.raises(PermissionError):
        await zigpy.serial.create_serial_connection(
            loop, protocol_factory, url=denied_port
        )

    # Locked
    locked_port = tmp_path / "locked"
    with locked_port.open("w") as f:
        # Lock the file
        fcntl.flock(f, fcntl.LOCK_EX | fcntl.LOCK_NB)

        with pytest.raises(
            PermissionError, match="The serial port is locked by another application"
        ):
            await zigpy.serial.create_serial_connection(
                loop, protocol_factory, url=locked_port
            )


async def test_serial_protocol() -> None:
    class SampleSerialProtocol(zigpy.serial.SerialProtocol):
        pass

    loop = asyncio.get_running_loop()

    protocol = SampleSerialProtocol()

    transport = Mock()
    loop.call_soon(protocol.connection_made, transport)

    # Connect
    await protocol.wait_until_connected()

    # Receive some data
    protocol.data_received(b"Hello")
    protocol.data_received(b" ")
    protocol.data_received(b"world")
    assert protocol._buffer == b"Hello world"

    # Close the transport
    asyncio.get_event_loop().call_soon(protocol.connection_lost, None)
    await protocol.disconnect()
    assert transport.close.mock_calls == [call()]



"""Works With All Hubs"""

from __future__ import annotations

from typing import Final

import zigpy.types as t
from zigpy.zcl import Cluster
from zigpy.zcl.foundation import (
    ZCL_CLUSTER_REVISION_ATTR,
    ZCL_REPORTING_STATUS_ATTR,
    BaseAttributeDefs,
    BaseCommandDefs,
    Status,
    ZCLAttributeDef,
    ZCLCommandDef,
)


class WwahIasZoneEnrollmentMode(t.enum8):
    """IAS Zone Enrollment Mode."""

    TripToPair = 0x00
    AutoEnrollmentResponse = 0x01
    Request = 0x02


class WwahPowerNotificationReason(t.enum8):
    """Power Notification Reason."""

    Unknown = 0x00
    Battery = 0x01
    Brownout = 0x02
    Watchdog = 0x03
    ResetPin = 0x04
    MemoryHardwareFault = 0x05
    SoftwareException = 0x06  # Spelled `SofwareException` in the SDK
    OtaBootloadSuccess = 0x07
    SoftwareReset = 0x08
    PowerButton = 0x09
    Temperature = 0x0A
    BootloadFailure = 0x0B


class WwahBeaconSurvey(t.Struct):
    """WWAH Beacon Survey structure."""

    device_short: t.uint16_t
    rssi: t.uint8_t
    classification_mask: t.uint8_t


class WwahClusterStatusToUseTC(t.Struct):
    """WWAH Cluster Status to Use TC structure."""

    cluster_id: t.ClusterId
    status: Status


# This cluster uses the Amazon manufacturer code for all commands and attributes
class WorksWithAllHubs(Cluster):
    """Works With All Hubs cluster"""

    cluster_id: Final[t.uint16_t] = 0xFC57
    ep_attribute: Final = "works_with_all_hubs"

    class AttributeDefs(BaseAttributeDefs):
        # 0x0000 and 0x0001 were removed from the spec
        disable_ota_downgrades: Final = ZCLAttributeDef(
            id=0x0002,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        mgmt_leave_without_rejoin_enabled: Final = ZCLAttributeDef(
            id=0x0003,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        nwk_retry_count: Final = ZCLAttributeDef(
            id=0x0004,
            type=t.uint8_t,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        mac_retry_count: Final = ZCLAttributeDef(
            id=0x0005,
            type=t.uint8_t,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        router_checkin_enabled: Final = ZCLAttributeDef(
            id=0x0006,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        touchlink_interpan_enabled: Final = ZCLAttributeDef(
            id=0x0007,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        wwah_parent_classification_enabled: Final = ZCLAttributeDef(
            id=0x0008,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        wwah_app_event_retry_enabled: Final = ZCLAttributeDef(
            id=0x0009,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        wwah_app_event_retry_queue_size: Final = ZCLAttributeDef(
            id=0x000A,
            type=t.uint8_t,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        wwah_rejoin_enabled: Final = ZCLAttributeDef(
            id=0x000B,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        mac_poll_failure_wait_time: Final = ZCLAttributeDef(
            id=0x000C,
            type=t.uint8_t,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        configuration_mode_enabled: Final = ZCLAttributeDef(
            id=0x000D,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        current_debug_report_id: Final = ZCLAttributeDef(
            id=0x000E,
            type=t.uint8_t,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        tc_security_on_ntwk_key_rotation_enabled: Final = ZCLAttributeDef(
            id=0x000F,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        wwah_bad_parent_recovery_enabled: Final = ZCLAttributeDef(
            id=0x0010,
            type=t.Bool,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        pending_network_update_channel: Final = ZCLAttributeDef(
            id=0x0011,
            type=t.uint8_t,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        pending_network_update_panid: Final = ZCLAttributeDef(
            id=0x0012,
            type=t.PanId,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        ota_max_offline_duration: Final = ZCLAttributeDef(
            id=0x0013,
            type=t.uint16_t,
            access="r",
            mandatory=False,
            manufacturer_code=0x1217,
        )
        cluster_revision: Final = ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = ZCL_REPORTING_STATUS_ATTR

    class ServerCommandDefs(BaseCommandDefs):
        # Client-to-Server commands (sent by client, received by server)
        # Enable enforcement of APS-level security for all cluster commands.
        enable_aps_link_key_authorization: Final = ZCLCommandDef(
            id=0x00,
            schema={"cluster_id": t.LVList[t.ClusterId, t.uint8_t]},
            manufacturer_code=0x1217,
        )

        # Disable enforcement of APS-level security for all cluster commands.
        disable_aps_link_key_authorization: Final = ZCLCommandDef(
            id=0x01,
            schema={"cluster_id": t.LVList[t.ClusterId, t.uint8_t]},
            manufacturer_code=0x1217,
        )

        # Query status of APS-level security enforcement for a specified cluster.
        aps_link_key_authorization_query: Final = ZCLCommandDef(
            id=0x02,
            schema={"cluster_id": t.ClusterId},
            manufacturer_code=0x1217,
        )

        # Trigger device to request a new APS link key from the Trust Center.
        request_new_aps_link_key: Final = ZCLCommandDef(
            id=0x03,
            schema={},
            manufacturer_code=0x1217,
        )

        enable_wwah_app_event_retry_algorithm: Final = ZCLCommandDef(
            id=0x04,
            schema={
                "first_backoff_time_seconds": t.uint8_t,
                "backoff_seq_common_ratio": t.uint8_t,
                "max_backoff_time_seconds": t.uint32_t,
                "max_redelivery_attempts": t.uint8_t,
            },
            manufacturer_code=0x1217,
        )

        disable_wwah_app_event_retry_algorithm: Final = ZCLCommandDef(
            id=0x05,
            schema={},
            manufacturer_code=0x1217,
        )

        # Trigger device to request current attribute values from Time Cluster server.
        request_time: Final = ZCLCommandDef(
            id=0x06,
            schema={},
            manufacturer_code=0x1217,
        )

        enable_wwah_rejoin_algorithm: Final = ZCLCommandDef(
            id=0x07,
            schema={
                "fast_rejoin_timeout_seconds": t.uint16_t,
                "duration_between_rejoins_seconds": t.uint16_t,
                "fast_rejoin_first_backoff_seconds": t.uint16_t,
                "max_backoff_time_seconds": t.uint16_t,
                "max_backoff_iterations": t.uint16_t,
            },
            manufacturer_code=0x1217,
        )

        disable_wwah_rejoin_algorithm: Final = ZCLCommandDef(
            id=0x08,
            schema={},
            manufacturer_code=0x1217,
        )

        # Set the enrollment method of an IAS Zone server.
        set_ias_zone_enrollment_method: Final = ZCLCommandDef(
            id=0x09,
            schema={"enrollment_mode": WwahIasZoneEnrollmentMode},
            manufacturer_code=0x1217,
        )

        clear_binding_table: Final = ZCLCommandDef(
            id=0x0A,
            schema={},
            manufacturer_code=0x1217,
        )

        # Enable device to periodically check connectivity with Zigbee Coordinator.
        enable_periodic_router_check_ins: Final = ZCLCommandDef(
            id=0x0B,
            schema={"check_in_interval": t.uint16_t},
            manufacturer_code=0x1217,
        )

        # Disable device from periodically checking connectivity with Zigbee Coordinator.
        disable_periodic_router_check_ins: Final = ZCLCommandDef(
            id=0x0C,
            schema={},
            manufacturer_code=0x1217,
        )

        set_mac_poll_failure_wait_time: Final = ZCLCommandDef(
            id=0x0D,
            schema={"wait_time": t.uint8_t},
            manufacturer_code=0x1217,
        )

        # Set pending network update parameters.
        set_pending_network_update: Final = ZCLCommandDef(
            id=0x0E,
            schema={"channel": t.uint8_t, "pan_id": t.PanId},
            manufacturer_code=0x1217,
        )

        # Require all unicast commands to have APS ACKs enabled.
        require_aps_acks_on_unicasts: Final = ZCLCommandDef(
            id=0x0F,
            schema={"cluster_id": t.LVList[t.ClusterId, t.uint8_t]},
            manufacturer_code=0x1217,
        )

        # Roll back changes made by Require APS ACK on Unicasts.
        remove_aps_acks_on_unicasts_requirement: Final = ZCLCommandDef(
            id=0x10,
            schema={},
            manufacturer_code=0x1217,
        )

        # Query whether unicast commands are required to have APS ACKs enabled.
        aps_ack_requirement_query: Final = ZCLCommandDef(
            id=0x11,
            schema={},
            manufacturer_code=0x1217,
        )

        debug_report_query: Final = ZCLCommandDef(
            id=0x12,
            schema={"debug_report_id": t.uint8_t},
            manufacturer_code=0x1217,
        )

        # Causes device to perform a scan for beacons advertising the device's network.
        survey_beacons: Final = ZCLCommandDef(
            id=0x13,
            schema={"standard_beacons": t.Bool},
            manufacturer_code=0x1217,
        )

        # Disallow OTA downgrade of all device firmware components.
        disable_ota_downgrades: Final = ZCLCommandDef(
            id=0x14,
            schema={},
            manufacturer_code=0x1217,
        )

        # Causes device to ignore MGMT Leave Without Rejoin commands.
        disable_mgmt_leave_without_rejoin: Final = ZCLCommandDef(
            id=0x15,
            schema={},
            manufacturer_code=0x1217,
        )

        # Causes device to ignore Touchlink Interpan messages.
        disable_touchlink_interpan_message_support: Final = ZCLCommandDef(
            id=0x16,
            schema={},
            manufacturer_code=0x1217,
        )

        enable_wwah_parent_classification: Final = ZCLCommandDef(
            id=0x17,
            schema={},
            manufacturer_code=0x1217,
        )

        disable_wwah_parent_classification: Final = ZCLCommandDef(
            id=0x18,
            schema={},
            manufacturer_code=0x1217,
        )

        # Process only network key rotation commands sent via unicast and encrypted by Trust Center Link Key.
        enable_tc_security_on_ntwk_key_rotation: Final = ZCLCommandDef(
            id=0x19,
            schema={},
            manufacturer_code=0x1217,
        )

        enable_wwah_bad_parent_recovery: Final = ZCLCommandDef(
            id=0x1A,
            schema={},
            manufacturer_code=0x1217,
        )

        disable_wwah_bad_parent_recovery: Final = ZCLCommandDef(
            id=0x1B,
            schema={},
            manufacturer_code=0x1217,
        )

        enable_configuration_mode: Final = ZCLCommandDef(
            id=0x1C,
            schema={},
            manufacturer_code=0x1217,
        )

        disable_configuration_mode: Final = ZCLCommandDef(
            id=0x1D,
            schema={},
            manufacturer_code=0x1217,
        )

        # Use only the Trust Center as cluster server for the set of clusters specified.
        use_trust_center_for_cluster_server: Final = ZCLCommandDef(
            id=0x1E,
            schema={"cluster_id": t.LVList[t.ClusterId, t.uint8_t]},
            manufacturer_code=0x1217,
        )

        # Causes device to send an appropriate Trust Center for Cluster Server Query Response command.
        trust_center_for_cluster_server_query: Final = ZCLCommandDef(
            id=0x1F,
            schema={},
            manufacturer_code=0x1217,
        )

    class ClientCommandDefs(BaseCommandDefs):
        aps_link_key_authorization_query_response: Final = ZCLCommandDef(
            id=0x00,
            schema={"cluster_id": t.ClusterId, "aps_link_key_auth_status": t.Bool},
            manufacturer_code=0x1217,
        )

        powering_off_notification: Final = ZCLCommandDef(
            id=0x01,
            schema={
                "power_notification_reason": WwahPowerNotificationReason,
                "manufacturer_id": t.uint16_t,
                "manufacturer_reason": t.LVList[t.uint8_t, t.uint8_t],
            },
            manufacturer_code=0x1217,
        )

        powering_on_notification: Final = ZCLCommandDef(
            id=0x02,
            schema={
                "power_notification_reason": WwahPowerNotificationReason,
                "manufacturer_id": t.uint16_t,
                "manufacturer_reason": t.LVList[t.uint8_t, t.uint8_t],
            },
            manufacturer_code=0x1217,
        )

        short_address_change: Final = ZCLCommandDef(
            id=0x03,
            schema={"device_eui64": t.EUI64, "device_short": t.uint16_t},
            manufacturer_code=0x1217,
        )

        aps_ack_enablement_query_response: Final = ZCLCommandDef(
            id=0x04,
            schema={"cluster_id": t.LVList[t.ClusterId, t.uint8_t]},
            manufacturer_code=0x1217,
        )

        power_descriptor_change: Final = ZCLCommandDef(
            id=0x05,
            schema={
                "current_power_mode": t.uint32_t,
                "available_power_sources": t.uint32_t,
                "current_power_source": t.uint32_t,
                "current_power_source_level": t.uint32_t,
            },
            manufacturer_code=0x1217,
        )

        new_debug_report_notification: Final = ZCLCommandDef(
            id=0x06,
            schema={"debug_report_id": t.uint8_t, "debug_report_size": t.uint32_t},
            manufacturer_code=0x1217,
        )

        debug_report_query_response: Final = ZCLCommandDef(
            id=0x07,
            schema={
                "debug_report_id": t.uint8_t,
                "debug_report_data": t.LVList[t.uint8_t, t.uint8_t],
            },
            manufacturer_code=0x1217,
        )

        trust_center_for_cluster_server_query_response: Final = ZCLCommandDef(
            id=0x08,
            schema={"cluster_id": t.LVList[t.ClusterId, t.uint8_t]},
            manufacturer_code=0x1217,
        )

        survey_beacons_response: Final = ZCLCommandDef(
            id=0x09,
            schema={"beacon": t.LVList[WwahBeaconSurvey, t.uint8_t]},
            manufacturer_code=0x1217,
        )

        # Whether this command exists on real devices or not is uncertain, the Silicon
        # Labs SDK XML made it sound like it was either not yet added or is being
        # removed
        use_trust_center_for_cluster_server_response: Final = ZCLCommandDef(
            id=0x9E,
            schema={
                "status": Status,
                "cluster_status": t.LVList[WwahClusterStatusToUseTC, t.uint8_t],
            },
            manufacturer_code=0x1217,
        )



"""OTA Firmware providers."""

from __future__ import annotations

import asyncio
import dataclasses
import datetime
import hashlib
import json
import logging
import pathlib
import re
import ssl
import typing
import urllib.parse

import aiohttp
import attrs
import jsonschema
import voluptuous as vol

import zigpy.config
from zigpy.ota import json_schemas
from zigpy.ota.image import BaseOTAImage, parse_ota_image
import zigpy.types as t
import zigpy.util

if typing.TYPE_CHECKING:
    import zigpy.device

LOGGER = logging.getLogger(__name__)

# `openssl s_client -connect fw.ota.homesmart.ikea.com:443 -showcerts`
IKEA_CERT = """\
-----BEGIN CERTIFICATE-----
MIICGDCCAZ+gAwIBAgIUdfH0KDnENv/dEcxH8iVqGGGDqrowCgYIKoZIzj0EAwMw
SzELMAkGA1UEBhMCU0UxGjAYBgNVBAoMEUlLRUEgb2YgU3dlZGVuIEFCMSAwHgYD
VQQDDBdJS0VBIEhvbWUgc21hcnQgUm9vdCBDQTAgFw0yMTA1MjYxOTAxMDlaGA8y
MDcxMDUxNDE5MDEwOFowSzELMAkGA1UEBhMCU0UxGjAYBgNVBAoMEUlLRUEgb2Yg
U3dlZGVuIEFCMSAwHgYDVQQDDBdJS0VBIEhvbWUgc21hcnQgUm9vdCBDQTB2MBAG
ByqGSM49AgEGBSuBBAAiA2IABIDRUvKGFMUu2zIhTdgfrfNcPULwMlc0TGSrDLBA
oTr0SMMV4044CRZQbl81N4qiuHGhFzCnXapZogkiVuFu7ZqSslsFuELFjc6ZxBjk
Kmud+pQM6QQdsKTE/cS06dA+P6NCMEAwDwYDVR0TAQH/BAUwAwEB/zAdBgNVHQ4E
FgQUcdlEnfX0MyZA4zAdY6CLOye9wfwwDgYDVR0PAQH/BAQDAgGGMAoGCCqGSM49
BAMDA2cAMGQCMG6mFIeB2GCFch3r0Gre4xRH+f5pn/bwLr9yGKywpeWvnUPsQ1KW
ckMLyxbeNPXdQQIwQc2YZDq/Mz0mOkoheTUWiZxK2a5bk0Uz1XuGshXmQvEg5TGy
2kVHW/Mz9/xwpy4u
-----END CERTIFICATE-----"""

# `openssl s_client -connect otau.meethue.com:443 -showcerts`
HUE_CERT = """\
-----BEGIN CERTIFICATE-----
MIIBwDCCAWagAwIBAgIJAJtrMkoTxs+WMAoGCCqGSM49BAMCMDIxCzAJBgNVBAYT
Ak5MMRQwEgYDVQQKDAtQaGlsaXBzIEh1ZTENMAsGA1UEAwwEcm9vdDAgFw0xNjA4
MjUwNzU5NDNaGA8yMDY4MDEwNTA3NTk0M1owMjELMAkGA1UEBhMCTkwxFDASBgNV
BAoMC1BoaWxpcHMgSHVlMQ0wCwYDVQQDDARyb290MFkwEwYHKoZIzj0CAQYIKoZI
zj0DAQcDQgAEENC1JOl6BxJrwCb+YK655zlM57VKFSi5OHDsmlCaF/EfTGGgU08/
JUtkCyMlHUUoYBZyzCBKXqRKkrT512evEKNjMGEwHQYDVR0OBBYEFAlkFYACVzir
qTr++cWia8AKH/fOMB8GA1UdIwQYMBaAFAlkFYACVzirqTr++cWia8AKH/fOMA8G
A1UdEwEB/wQFMAMBAf8wDgYDVR0PAQH/BAQDAgGGMAoGCCqGSM49BAMCA0gAMEUC
IQDcGfyXaUl5hjr5YE8m2piXhMcDzHTNbO1RvGgz4r9IswIgFTTw/R85KyfIiW+E
clwJRVSsq8EApeFREenCkRM0EIk=
-----END CERTIFICATE-----"""

OTA_PROVIDER_TYPES: dict[str, type[BaseOtaProvider]] = {}


def register_provider(provider: type[BaseOtaProvider]) -> type[BaseOtaProvider]:
    """Register a new OTA provider."""
    OTA_PROVIDER_TYPES[provider.NAME] = provider
    return provider


@attrs.define(frozen=True, kw_only=True)
class BaseOtaImageMetadata(t.BaseDataclassMixin):
    file_version: int
    manufacturer_id: int | None = None
    image_type: int | None = None

    checksum: str | None = None
    file_size: int | None = None

    manufacturer_names: tuple[str, ...] = ()
    model_names: tuple[str, ...] = ()

    changelog: str | None = None
    release_notes: str | None = None
    release_url: str | None = None

    min_hardware_version: int | None = None
    max_hardware_version: int | None = None
    min_current_file_version: int | None = None
    max_current_file_version: int | None = None
    specificity: int | None = None

    source: str = "Unknown"
    trusted: bool = False

    async def _fetch(self) -> bytes:
        raise NotImplementedError

    async def _validate(self, data: bytes) -> None:
        if self.file_size is not None and len(data) != self.file_size:
            raise ValueError(
                f"Image size is invalid: expected {self.file_size} bytes,"
                f" got {len(data)} bytes"
            )

        if self.checksum is not None:
            algorithm, checksum = self.checksum.split(":")
            hasher = hashlib.new(algorithm)
            await asyncio.get_running_loop().run_in_executor(None, hasher.update, data)

            if hasher.hexdigest() != checksum:
                raise ValueError(
                    f"Image checksum is invalid: expected {checksum},"
                    f" got {hasher.hexdigest()}"
                )

    async def fetch(self) -> BaseOTAImage:
        data = await self._fetch()
        await self._validate(data)

        image, _ = parse_ota_image(data)
        return image


@attrs.define(frozen=True, kw_only=True)
class RemoteOtaImageMetadata(BaseOtaImageMetadata):
    url: str

    # If a provider uses a self-signed certificate, it can override this
    ssl_ctx: ssl.SSLContext | None = None

    async def _fetch(self) -> bytes:
        async with aiohttp.ClientSession(raise_for_status=True) as req:
            async with req.get(self.url, ssl=self.ssl_ctx) as rsp:
                return await rsp.read()


@attrs.define(frozen=True, kw_only=True)
class LocalOtaImageMetadata(BaseOtaImageMetadata):
    path: pathlib.Path

    async def _fetch(self) -> bytes:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self.path.read_bytes)


@attrs.define(frozen=True, kw_only=True)
class IkeaRemoteOtaImageMetadata(RemoteOtaImageMetadata):
    ssl_ctx = dataclasses.field(default_factory=lambda: Tradfri.SSL_CTX)

    async def _fetch(self) -> bytes:
        async with aiohttp.ClientSession(raise_for_status=True) as req:
            # Use IKEA's self-signed certificate
            async with req.get(self.url, ssl=Tradfri.SSL_CTX) as rsp:
                return await rsp.read()


@attrs.define(frozen=True, kw_only=True)
class SignedIkeaRemoteOtaImageMetadata(IkeaRemoteOtaImageMetadata):
    ssl_ctx = dataclasses.field(default_factory=lambda: Tradfri.SSL_CTX)

    async def _validate(self, data: bytes) -> None:
        ota_offset = int.from_bytes(data[16:20], "little")
        ota_size = int.from_bytes(data[20:24], "little")
        block_size = int.from_bytes(data[32:36], "little")
        num_block_hashes = int.from_bytes(data[36:40], "little")

        if (
            not data.startswith(b"NGIS")
            or self.file_size != ota_size
            or 40 + 32 * num_block_hashes != ota_offset
            or block_size * num_block_hashes < ota_size
        ):
            raise ValueError(f"Invalid signed container: {data[:16]!r}")

        loop = asyncio.get_running_loop()

        for block_num in range(num_block_hashes):
            offset = ota_offset + block_size * block_num
            size = block_size - max(0, offset + block_size - (ota_offset + ota_size))

            block = data[offset : offset + size]
            expected_checksum = data[40 + 32 * block_num : 40 + 32 * (block_num + 1)]
            hasher = await loop.run_in_executor(None, hashlib.sha256, block)

            if hasher.digest() != expected_checksum:
                raise ValueError(f"Block {block_num} has invalid checksum")


class BaseOtaProvider:
    NAME: str
    MANUFACTURER_IDS: tuple[int, ...] = ()
    DEFAULT_URL: str | None = None
    VOL_SCHEMA: vol.Schema
    JSON_SCHEMA: dict | None = None
    INDEX_EXPIRATION_TIME = datetime.timedelta(hours=24)
    TRUSTED: bool = False

    def __init__(
        self,
        url: str | typing.Literal[True] | None = None,
        manufacturer_ids: list[int] | None = None,
        *,
        override_previous: bool = False,
    ) -> None:
        self.url = self.DEFAULT_URL if url in (True, None) else url
        self._index_last_updated = datetime.datetime.fromtimestamp(0, tz=datetime.UTC)

        if manufacturer_ids is not None:
            self.manufacturer_ids = tuple(manufacturer_ids)
        else:
            self.manufacturer_ids = tuple(self.MANUFACTURER_IDS)

        self.override_previous = override_previous

    def compatible_with_device(self, device: zigpy.device.Device) -> bool:
        if not self.manufacturer_ids:
            return True

        return device.manufacturer_id in self.manufacturer_ids

    async def load_index(self) -> list[BaseOtaImageMetadata] | None:
        now = datetime.datetime.now(datetime.UTC)

        # Don't hammer the OTA indexes too frequently
        if now - self._index_last_updated < self.INDEX_EXPIRATION_TIME:
            return None

        try:
            async with aiohttp.ClientSession(
                headers={"accept": "application/json"},
                raise_for_status=True,
            ) as session:
                return [meta async for meta in self._load_index(session)]
        finally:
            self._index_last_updated = now

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        if typing.TYPE_CHECKING:
            yield

        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseOtaProvider):
            return NotImplemented

        # Providers of a different concrete type are never equal
        if type(self) is not type(other):
            return False

        return self.url == other.url and self.manufacturer_ids == other.manufacturer_ids

    def __hash__(self) -> int:
        return hash((type(self), self.url, self.manufacturer_ids))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(url={self.url!r}, manufacturer_ids={self.manufacturer_ids!r})"


@register_provider
class Tradfri(BaseOtaProvider):
    NAME = "ikea"
    MANUFACTURER_IDS = (4476,)
    DEFAULT_URL = "https://fw.ota.homesmart.ikea.com/DIRIGERA/version_info.json"
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_URL
    JSON_SCHEMA = json_schemas.TRADFRI_SCHEMA

    SSL_CTX: ssl.SSLContext = ssl.create_default_context()
    SSL_CTX.load_verify_locations(cadata=IKEA_CERT)

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        async with session.get(self.url, ssl=self.SSL_CTX) as rsp:
            # IKEA does not always respond with an appropriate Content-Type but the
            # response is always JSON
            fw_lst = await rsp.json(content_type=None)

        jsonschema.validate(fw_lst, self.JSON_SCHEMA)

        for fw in fw_lst:
            # Skip the gateway image
            if "fw_image_type" not in fw:
                continue

            if "fw_sha3_256" in fw:
                # New style IKEA
                file_version_match = re.match(r".*_v(?P<v>\d+)_.*", fw["fw_binary_url"])

                if file_version_match is None:
                    LOGGER.warning("Could not parse IKEA OTA JSON: %r", fw)
                    continue

                image = IkeaRemoteOtaImageMetadata(
                    file_version=int(file_version_match.group("v"), 10),
                    manufacturer_id=self.MANUFACTURER_IDS[0],
                    image_type=fw["fw_image_type"],
                    checksum="sha3-256:" + fw["fw_sha3_256"],
                    url=fw["fw_binary_url"],
                    source="IKEA (DIRIGERA)",
                )
            else:
                # Old style IKEA
                if fw["fw_type"] != 2:
                    continue

                image = SignedIkeaRemoteOtaImageMetadata(
                    file_version=(
                        (fw["fw_file_version_MSB"] << 16)
                        | (fw["fw_file_version_LSB"] << 0)
                    ),
                    manufacturer_id=fw["fw_manufacturer_id"],
                    image_type=fw["fw_image_type"],
                    # The file size is of the contained image, not the container!
                    file_size=fw["fw_filesize"],
                    url=fw["fw_binary_url"].replace("http://", "https://", 1),
                    source="IKEA (TRÅDFRI)",
                )

            # Bricking update: https://github.com/zigpy/zigpy/issues/1428
            if image.image_type in (8704, 8710):
                continue

            yield image


@register_provider
class Ledvance(BaseOtaProvider):
    NAME = "ledvance"
    # This isn't static but no more than these two have ever existed
    MANUFACTURER_IDS = (4489, 4364)
    DEFAULT_URL = "https://api.update.ledvance.com/v1/zigbee/firmwares"
    JSON_SCHEMA = json_schemas.LEDVANCE_SCHEMA
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_URL

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        async with session.get(self.url) as rsp:
            fw_lst = await rsp.json()

        jsonschema.validate(fw_lst, self.JSON_SCHEMA)

        for fw in fw_lst["firmwares"]:
            identity = fw["identity"]
            version = identity["version"]

            yield RemoteOtaImageMetadata(
                file_version=int(fw["fullName"].split("/")[1], 16),
                manufacturer_id=identity["company"],
                image_type=identity["product"],
                checksum="sha256:" + fw["shA256"],
                file_size=fw["length"],
                model_names=(fw["productName"],),
                url=(
                    "https://api.update.ledvance.com/v1/zigbee/firmwares/download?"
                    + urllib.parse.urlencode(
                        {
                            "Company": identity["company"],
                            "Product": identity["product"],
                            "Version": (
                                f"{version['major']}.{version['minor']}"
                                f".{version['build']}.{version['revision']}"
                            ),
                        }
                    )
                ),
                release_notes=fw["releaseNotes"],
                source="Ledvance",
            )


class StubOtaProvider(BaseOtaProvider):
    """Stub provider to keep existing configurations working."""

    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_URL

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        if typing.TYPE_CHECKING:
            yield  # pragma: no cover


@register_provider
class Salus(StubOtaProvider):
    NAME = "salus"
    MANUFACTURER_IDS = (4216, 43981)


@register_provider
class ThirdReality(StubOtaProvider):
    NAME = "thirdreality"
    MANUFACTURER_IDS = (4659, 4877, 5127)


@register_provider
class Inovelli(StubOtaProvider):
    NAME = "inovelli"
    MANUFACTURER_IDS = (4655,)


@register_provider
class Sonoff(BaseOtaProvider):
    NAME = "sonoff"
    MANUFACTURER_IDS = (4742,)

    JSON_SCHEMA = json_schemas.SONOFF_SCHEMA
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_URL

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        async with session.get(
            "https://zigbee-ota.sonoff.tech/releases/upgrade.json"
        ) as rsp:
            fw_lst = await rsp.json()

        jsonschema.validate(fw_lst, self.JSON_SCHEMA)

        for fw in fw_lst:
            yield RemoteOtaImageMetadata(
                file_version=fw["fw_file_version"],
                manufacturer_id=fw["fw_manufacturer_id"],
                image_type=fw["fw_image_type"],
                file_size=fw["fw_filesize"],
                url=fw["fw_binary_url"],
                model_names=(fw["model_id"],),
                source="Sonoff",
            )


class BaseZigpyProvider(BaseOtaProvider):
    JSON_SCHEMA = json_schemas.REMOTE_PROVIDER_SCHEMA

    @classmethod
    def _load_zigpy_index(
        cls,
        index: dict,
        *,
        index_root: pathlib.Path | None = None,
        ssl_ctx: ssl.SSLContext | None = None,
    ):
        jsonschema.validate(index, cls.JSON_SCHEMA)

        for fw in index["firmwares"]:
            shared_kwargs = {
                "file_version": fw["file_version"],
                "manufacturer_id": fw["manufacturer_id"],
                "image_type": fw["image_type"],
                "manufacturer_names": tuple(fw.get("manufacturer_names", [])),
                "model_names": tuple(fw.get("model_names", [])),
                "checksum": fw["checksum"],
                "file_size": fw["file_size"],
                "min_hardware_version": fw.get("min_hardware_version"),
                "max_hardware_version": fw.get("max_hardware_version"),
                "min_current_file_version": fw.get("min_current_file_version"),
                "max_current_file_version": fw.get("max_current_file_version"),
                "changelog": fw.get("changelog"),
                "release_notes": fw.get("release_notes"),
                "release_url": fw.get("release_url"),
                "specificity": fw.get("specificity"),
                "source": "",  # Set in a subclass
            }

            if "path" in fw and index_root is not None:
                yield LocalOtaImageMetadata(
                    **shared_kwargs, path=index_root / fw["path"]
                )
            else:
                yield RemoteOtaImageMetadata(
                    **shared_kwargs, url=fw["binary_url"], ssl_ctx=ssl_ctx
                )


@register_provider
class LocalZigpyProvider(BaseZigpyProvider):
    NAME = "zigpy_local"
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_JSON_INDEX

    def __init__(self, index_file: pathlib.Path, **kwargs):
        super().__init__(url=None, **kwargs)
        self.index_file = index_file

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        index_text = await asyncio.get_running_loop().run_in_executor(
            None, self.index_file.read_text
        )
        index = json.loads(index_text)

        for img in self._load_zigpy_index(index, index_root=self.index_file.parent):
            yield img.replace(source=f"Local zigpy provider ({self.index_file})")

    def __eq__(self, other: object) -> bool:
        if (
            not isinstance(other, self.__class__)
            or super().__eq__(other) is NotImplemented
        ):
            return NotImplemented

        return super().__eq__(other) and self.index_file == other.index_file

    def __hash__(self) -> int:
        return hash((type(self), self.index_file, self.manufacturer_ids))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(index_file={self.index_file!r}, manufacturer_ids={self.manufacturer_ids!r})"


@register_provider
class RemoteZigpyProvider(BaseZigpyProvider):
    NAME = "zigpy_remote"
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_URL_REQUIRED

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        async with session.get(self.url) as rsp:
            fw_lst = await rsp.json(content_type=None)

        jsonschema.validate(fw_lst, self.JSON_SCHEMA)

        for img in self._load_zigpy_index(fw_lst):
            yield img.replace(source=f"Remote zigpy provider ({self.url})")


class BaseZ2MProvider(BaseOtaProvider):
    JSON_SCHEMA = json_schemas.Z2M_SCHEMA

    @classmethod
    def _load_z2m_index(
        cls,
        index: dict,
        *,
        index_root: pathlib.Path | None = None,
        ssl_ctx: ssl.SSLContext | None = None,
    ) -> typing.Iterator[LocalOtaImageMetadata | RemoteOtaImageMetadata]:
        jsonschema.validate(index, cls.JSON_SCHEMA)

        for fw in index:
            shared_kwargs = {
                "file_version": fw["fileVersion"],
                "manufacturer_id": fw["manufacturerCode"],
                "image_type": fw["imageType"],
                "checksum": "sha512:" + fw["sha512"],
                "file_size": fw["fileSize"],
                "manufacturer_names": tuple(fw.get("manufacturerName", [])),
                "model_names": tuple([fw["modelId"]] if "modelId" in fw else []),
                "min_current_file_version": fw.get("minFileVersion"),
                "max_current_file_version": fw.get("maxFileVersion"),
                "min_hardware_version": fw.get("hardwareVersionMin"),
                "max_hardware_version": fw.get("hardwareVersionMax"),
                "changelog": fw.get("releaseNotes"),  # Changelog is short
                "source": "",  # Set in a subclass
            }

            if "path" in fw and index_root is not None:
                yield LocalOtaImageMetadata(
                    **shared_kwargs, path=index_root / fw["path"]
                )
            else:
                yield RemoteOtaImageMetadata(
                    **shared_kwargs, url=fw["url"], ssl_ctx=ssl_ctx
                )


@register_provider
class LocalZ2MProvider(BaseZ2MProvider):
    NAME = "z2m_local"
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_JSON_INDEX

    def __init__(self, index_file: pathlib.Path, **kwargs):
        super().__init__(**kwargs)
        self.index_file = index_file

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        index_text = await asyncio.get_running_loop().run_in_executor(
            None, self.index_file.read_text
        )
        index = json.loads(index_text)

        for img in self._load_z2m_index(index, index_root=self.index_file.parent):
            yield img.replace(source=f"Local Z2M provider ({self.index_file})")

    def __eq__(self, other: object) -> bool:
        if (
            not isinstance(other, self.__class__)
            or super().__eq__(other) is NotImplemented
        ):
            return NotImplemented

        return super().__eq__(other) and self.index_file == other.index_file

    def __hash__(self) -> int:
        return hash((type(self), self.index_file, self.manufacturer_ids))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(index_file={self.index_file!r}, manufacturer_ids={self.manufacturer_ids!r})"


@register_provider
class RemoteZ2MProvider(BaseZ2MProvider):
    NAME = "z2m"
    DEFAULT_URL = (
        "https://raw.githubusercontent.com/Koenkk/zigbee-OTA/master/index.json"
    )
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_URL

    SSL_CTX = ssl.create_default_context()
    SSL_CTX.load_verify_locations(cadata=HUE_CERT)

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        async with session.get(self.url) as rsp:
            fw_lst = await rsp.json(content_type=None)

        for img in self._load_z2m_index(fw_lst, ssl_ctx=self.SSL_CTX):
            yield img.replace(source=f"Remote Z2M provider ({self.url})")


@register_provider
class ZigpyOtaProvider(BaseZigpyProvider):
    """OTA provider for zigpy-ota repository with multiple release channels.

    The provider uses a two-step fetch process:
    1. Fetch version file from GitHub to get the current index URL
    2. Fetch the actual OTA index from the URL in the version file
    """

    NAME = "zigpy_ota"
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_ZIGPY_OTA
    TRUSTED = True

    DEFAULT_CHANNEL = "stable"
    SUPPORTED_CHANNELS = {"stable", "beta", "dev"}

    VERSION_FILE_BASE_URL = (
        "https://raw.githubusercontent.com/zigpy/zigpy-ota/release/version"
    )

    # Combined SSL context for downloading images from IKEA and Hue servers
    # (zigpy-ota index may point to images hosted on these servers)
    SSL_CTX: ssl.SSLContext = ssl.create_default_context()
    SSL_CTX.load_verify_locations(cadata=IKEA_CERT)
    SSL_CTX.load_verify_locations(cadata=HUE_CERT)

    def __init__(
        self,
        url: str | typing.Literal[True] | None = None,
        channel: str | None = None,
        **kwargs,
    ) -> None:
        # If a specific URL is provided, use it directly (for testing or custom endpoints)
        # Otherwise, construct the version file URL based on the channel
        if url in (True, None):
            self.channel = channel or self.DEFAULT_CHANNEL
            if self.channel not in self.SUPPORTED_CHANNELS:
                raise ValueError(
                    f"Invalid channel '{self.channel}'. Must be one of: {self.SUPPORTED_CHANNELS}"
                )
            url = f"{self.VERSION_FILE_BASE_URL}/{self.channel}.json"
        else:
            # Custom URL provided (direct to version file)
            self.channel = channel

        super().__init__(url=url, **kwargs)

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        # First fetch the version file to get the actual index URL
        async with session.get(self.url) as rsp:
            version_data = await rsp.json(content_type=None)

        # Extract the index URL from the version file
        # Format: {"schemas": {"zigpy_v2": {"version": "...", "url": "..."}}}
        index_url = version_data["schemas"]["zigpy_v2"]["url"]

        # Now fetch the actual OTA index
        async with session.get(index_url) as rsp:
            fw_lst = await rsp.json(content_type=None)

        jsonschema.validate(fw_lst, self.JSON_SCHEMA)

        for img in self._load_zigpy_index(fw_lst, ssl_ctx=self.SSL_CTX):
            channel_name = self.channel or "custom"
            yield img.replace(source=f"zigpy-ota provider ({channel_name} channel)")

    def __eq__(self, other: object) -> bool:
        if (
            not isinstance(other, self.__class__)
            or super().__eq__(other) is NotImplemented
        ):
            return NotImplemented

        return super().__eq__(other) and self.channel == other.channel

    def __hash__(self) -> int:
        return hash((type(self), self.url, self.channel, self.manufacturer_ids))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(url={self.url!r}, channel={self.channel!r}, manufacturer_ids={self.manufacturer_ids!r})"


@register_provider
class AdvancedFileProvider(BaseOtaProvider):
    NAME = "advanced"
    VOL_SCHEMA = zigpy.config.SCHEMA_OTA_PROVIDER_FOLDER

    def __init__(self, path: pathlib.Path, **kwargs):
        # The `vol` schema passes through the `warning` key, which is unused
        kwargs.pop("warning", None)

        super().__init__(url=None, **kwargs)
        self.path = path

    async def _load_index(
        self, session: aiohttp.ClientSession
    ) -> typing.AsyncIterator[BaseOtaImageMetadata]:
        loop = asyncio.get_running_loop()

        paths = await loop.run_in_executor(None, self.path.rglob, "*")

        async for chunk in zigpy.util.async_iterate_in_chunks(paths, chunk_size=100):
            for path in chunk:
                if not path.is_file():
                    continue

                data = await loop.run_in_executor(None, path.read_bytes)

                try:
                    image, _ = parse_ota_image(data)
                except Exception as exc:  # noqa: BLE001
                    LOGGER.debug("Failed to parse image %s: %r", path, exc)
                    continue

                # This protects against images being swapped out in the local filesystem
                hasher = await loop.run_in_executor(None, hashlib.sha1, data)

                yield LocalOtaImageMetadata(
                    path=path,
                    file_version=image.header.file_version,
                    manufacturer_id=image.header.manufacturer_id,
                    image_type=image.header.image_type,
                    checksum="sha1:" + hasher.hexdigest(),
                    file_size=len(data),
                    min_hardware_version=image.header.minimum_hardware_version,
                    max_hardware_version=image.header.maximum_hardware_version,
                    source=f"Advanced file provider ({self.path})",
                )

    def __eq__(self, other: object) -> bool:
        if (
            not isinstance(other, self.__class__)
            or super().__eq__(other) is NotImplemented
        ):
            return NotImplemented

        return super().__eq__(other) and self.path == other.path

    def __hash__(self) -> int:
        return hash((type(self), self.path, self.manufacturer_ids))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path!r}, manufacturer_ids={self.manufacturer_ids!r})"



from __future__ import annotations

from typing import Final

import zigpy.types as t
from zigpy.zcl import Cluster, foundation
from zigpy.zcl.foundation import (
    BaseAttributeDefs,
    BaseCommandDefs,
    ZCLAttributeDef,
    ZCLCommandDef,
)


class ApplianceIdentification(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0B00
    name: Final = "Appliance Identification"
    ep_attribute: Final = "appliance_id"

    class AttributeDefs(BaseAttributeDefs):
        basic_identification: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint56_t, access="r", mandatory=True
        )
        company_name: Final = ZCLAttributeDef(
            id=0x0010, type=t.LimitedCharString(16), access="r"
        )
        company_id: Final = ZCLAttributeDef(id=0x0011, type=t.uint16_t, access="r")
        brand_name: Final = ZCLAttributeDef(
            id=0x0012, type=t.LimitedCharString(16), access="r"
        )
        brand_id: Final = ZCLAttributeDef(id=0x0013, type=t.uint16_t, access="r")
        model: Final = ZCLAttributeDef(id=0x0014, type=t.LimitedLVBytes(16), access="r")
        part_number: Final = ZCLAttributeDef(
            id=0x0015, type=t.LimitedLVBytes(16), access="r"
        )
        product_revision: Final = ZCLAttributeDef(
            id=0x0016, type=t.LimitedLVBytes(6), access="r"
        )
        software_revision: Final = ZCLAttributeDef(
            id=0x0017, type=t.LimitedLVBytes(6), access="r"
        )
        product_type_name: Final = ZCLAttributeDef(
            id=0x0018, type=t.LVBytesSize2, access="r"
        )
        product_type_id: Final = ZCLAttributeDef(id=0x0019, type=t.uint16_t, access="r")
        ceced_specification_version: Final = ZCLAttributeDef(
            id=0x001A, type=t.uint8_t, access="r"
        )
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class MeterIdentification(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0B01
    name: Final = "Meter Identification"
    ep_attribute: Final = "meter_id"

    class AttributeDefs(BaseAttributeDefs):
        company_name: Final = ZCLAttributeDef(
            id=0x0000, type=t.LimitedCharString(16), access="r", mandatory=True
        )
        meter_type_id: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        data_quality_id: Final = ZCLAttributeDef(
            id=0x0004, type=t.uint16_t, access="r", mandatory=True
        )
        customer_name: Final = ZCLAttributeDef(
            id=0x0005, type=t.LimitedCharString(16), access="rw"
        )
        model: Final = ZCLAttributeDef(id=0x0006, type=t.LimitedLVBytes(16), access="r")
        part_number: Final = ZCLAttributeDef(
            id=0x0007, type=t.LimitedLVBytes(16), access="r"
        )
        product_revision: Final = ZCLAttributeDef(
            id=0x0008, type=t.LimitedLVBytes(6), access="r"
        )
        software_revision: Final = ZCLAttributeDef(
            id=0x000A, type=t.LimitedLVBytes(6), access="r"
        )
        utility_name: Final = ZCLAttributeDef(
            id=0x000B, type=t.LimitedCharString(16), access="r"
        )
        pod: Final = ZCLAttributeDef(
            id=0x000C, type=t.LimitedCharString(16), access="r", mandatory=True
        )
        available_power: Final = ZCLAttributeDef(
            id=0x000D, type=t.int24s, access="r", mandatory=True
        )
        power_threshold: Final = ZCLAttributeDef(
            id=0x000E, type=t.int24s, access="r", mandatory=True
        )
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class ApplianceEventAlerts(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0B02
    name: Final = "Appliance Event Alerts"
    ep_attribute: Final = "appliance_event"

    class AttributeDefs(BaseAttributeDefs):
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR

    class ServerCommandDefs(BaseCommandDefs):
        get_alerts: Final = ZCLCommandDef(id=0x00, schema={})

    class ClientCommandDefs(BaseCommandDefs):
        get_alerts_response: Final = ZCLCommandDef(id=0x00, schema={})
        alerts_notification: Final = ZCLCommandDef(id=0x01, schema={})
        event_notification: Final = ZCLCommandDef(id=0x02, schema={})


class ApplianceStatistics(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0B03
    name: Final = "Appliance Statistics"
    ep_attribute: Final = "appliance_stats"

    class AttributeDefs(BaseAttributeDefs):
        log_max_size: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint32_t, access="r", mandatory=True
        )
        log_queue_max_size: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint8_t, access="r", mandatory=True
        )
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR

    class ServerCommandDefs(BaseCommandDefs):
        log: Final = ZCLCommandDef(id=0x00, schema={})
        log_queue: Final = ZCLCommandDef(id=0x01, schema={})

    class ClientCommandDefs(BaseCommandDefs):
        log_notification: Final = ZCLCommandDef(id=0x00, schema={})
        log_response: Final = ZCLCommandDef(id=0x01, schema={})
        log_queue_response: Final = ZCLCommandDef(id=0x02, schema={})
        statistics_available: Final = ZCLCommandDef(id=0x03, schema={})


class MeasurementType(t.bitmap32):
    Active_measurement_AC = 1 << 0
    Reactive_measurement_AC = 1 << 1
    Apparent_measurement_AC = 1 << 2
    Phase_A_measurement = 1 << 3
    Phase_B_measurement = 1 << 4
    Phase_C_measurement = 1 << 5
    DC_measurement = 1 << 6
    Harmonics_measurement = 1 << 7
    Power_quality_measurement = 1 << 8


class DCOverloadAlarmMark(t.bitmap8):
    Voltage_Overload = 1 << 0
    Current_Overload = 1 << 1


class ACAlarmsMask(t.bitmap16):
    Voltage_Overload = 1 << 0
    Current_Overload = 1 << 1
    Active_Power_Overload = 1 << 2
    Reactive_Power_Overload = 1 << 3
    Average_RMS_Over_Voltage = 1 << 4
    Average_RMS_Under_Voltage = 1 << 5
    RMS_Extreme_Over_Voltage = 1 << 6
    RMS_Extreme_Under_Voltage = 1 << 7
    RMS_Voltage_Sag = 1 << 8
    RMS_Voltage_Swell = 1 << 9


class ElectricalMeasurement(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0B04
    name: Final = "Electrical Measurement"
    ep_attribute: Final = "electrical_measurement"

    MeasurementType: Final = MeasurementType
    DCOverloadAlarmMark: Final = DCOverloadAlarmMark
    ACAlarmsMask: Final = ACAlarmsMask

    class AttributeDefs(BaseAttributeDefs):
        # Basic Information
        measurement_type: Final = ZCLAttributeDef(
            id=0x0000, type=MeasurementType, access="r", mandatory=True
        )
        # DC Measurement
        dc_voltage: Final = ZCLAttributeDef(id=0x0100, type=t.int16s, access="rp")
        dc_voltage_min: Final = ZCLAttributeDef(id=0x0101, type=t.int16s, access="r")
        dc_voltage_max: Final = ZCLAttributeDef(id=0x0102, type=t.int16s, access="r")
        dc_current: Final = ZCLAttributeDef(id=0x0103, type=t.int16s, access="rp")
        dc_current_min: Final = ZCLAttributeDef(id=0x0104, type=t.int16s, access="r")
        dc_current_max: Final = ZCLAttributeDef(id=0x0105, type=t.int16s, access="r")
        dc_power: Final = ZCLAttributeDef(id=0x0106, type=t.int16s, access="rp")
        dc_power_min: Final = ZCLAttributeDef(id=0x0107, type=t.int16s, access="r")
        dc_power_max: Final = ZCLAttributeDef(id=0x0108, type=t.int16s, access="r")
        # DC Formatting
        dc_voltage_multiplier: Final = ZCLAttributeDef(
            id=0x0200, type=t.uint16_t, access="rp"
        )
        dc_voltage_divisor: Final = ZCLAttributeDef(
            id=0x0201, type=t.uint16_t, access="rp"
        )
        dc_current_multiplier: Final = ZCLAttributeDef(
            id=0x0202, type=t.uint16_t, access="rp"
        )
        dc_current_divisor: Final = ZCLAttributeDef(
            id=0x0203, type=t.uint16_t, access="rp"
        )
        dc_power_multiplier: Final = ZCLAttributeDef(
            id=0x0204, type=t.uint16_t, access="rp"
        )
        dc_power_divisor: Final = ZCLAttributeDef(
            id=0x0205, type=t.uint16_t, access="rp"
        )
        # AC (Non-phase Specific) Measurements
        ac_frequency: Final = ZCLAttributeDef(id=0x0300, type=t.uint16_t, access="rp")
        ac_frequency_min: Final = ZCLAttributeDef(
            id=0x0301, type=t.uint16_t, access="r"
        )
        ac_frequency_max: Final = ZCLAttributeDef(
            id=0x0302, type=t.uint16_t, access="r"
        )
        neutral_current: Final = ZCLAttributeDef(
            id=0x0303, type=t.uint16_t, access="rp"
        )
        total_active_power: Final = ZCLAttributeDef(
            id=0x0304, type=t.int32s, access="rp"
        )
        total_reactive_power: Final = ZCLAttributeDef(
            id=0x0305, type=t.int32s, access="rp"
        )
        total_apparent_power: Final = ZCLAttributeDef(
            id=0x0306, type=t.uint32_t, access="rp"
        )
        meas1st_harmonic_current: Final = ZCLAttributeDef(
            id=0x0307, type=t.int16s, access="rp"
        )
        meas3rd_harmonic_current: Final = ZCLAttributeDef(
            id=0x0308, type=t.int16s, access="rp"
        )
        meas5th_harmonic_current: Final = ZCLAttributeDef(
            id=0x0309, type=t.int16s, access="rp"
        )
        meas7th_harmonic_current: Final = ZCLAttributeDef(
            id=0x030A, type=t.int16s, access="rp"
        )
        meas9th_harmonic_current: Final = ZCLAttributeDef(
            id=0x030B, type=t.int16s, access="rp"
        )
        meas11th_harmonic_current: Final = ZCLAttributeDef(
            id=0x030C, type=t.int16s, access="rp"
        )
        meas_phase1st_harmonic_current: Final = ZCLAttributeDef(
            id=0x030D, type=t.int16s, access="rp"
        )
        meas_phase3rd_harmonic_current: Final = ZCLAttributeDef(
            id=0x030E, type=t.int16s, access="rp"
        )
        meas_phase5th_harmonic_current: Final = ZCLAttributeDef(
            id=0x030F, type=t.int16s, access="rp"
        )
        meas_phase7th_harmonic_current: Final = ZCLAttributeDef(
            id=0x0310, type=t.int16s, access="rp"
        )
        meas_phase9th_harmonic_current: Final = ZCLAttributeDef(
            id=0x0311, type=t.int16s, access="rp"
        )
        meas_phase11th_harmonic_current: Final = ZCLAttributeDef(
            id=0x0312, type=t.int16s, access="rp"
        )
        # AC (Non-phase specific) Formatting
        ac_frequency_multiplier: Final = ZCLAttributeDef(
            id=0x0400, type=t.uint16_t, access="rp"
        )
        ac_frequency_divisor: Final = ZCLAttributeDef(
            id=0x0401, type=t.uint16_t, access="rp"
        )
        power_multiplier: Final = ZCLAttributeDef(
            id=0x0402, type=t.uint32_t, access="rp"
        )
        power_divisor: Final = ZCLAttributeDef(id=0x0403, type=t.uint32_t, access="rp")
        harmonic_current_multiplier: Final = ZCLAttributeDef(
            id=0x0404, type=t.int8s, access="rp"
        )
        phase_harmonic_current_multiplier: Final = ZCLAttributeDef(
            id=0x0405, type=t.int8s, access="rp"
        )
        # AC (Single Phase or Phase A) Measurements
        instantaneous_voltage: Final = ZCLAttributeDef(
            id=0x0500, type=t.int16s, access="rp"
        )
        instantaneous_line_current: Final = ZCLAttributeDef(
            id=0x0501, type=t.uint16_t, access="rp"
        )
        instantaneous_active_current: Final = ZCLAttributeDef(
            id=0x0502, type=t.int16s, access="rp"
        )
        instantaneous_reactive_current: Final = ZCLAttributeDef(
            id=0x0503, type=t.int16s, access="rp"
        )
        instantaneous_power: Final = ZCLAttributeDef(
            id=0x0504, type=t.int16s, access="rp"
        )
        rms_voltage: Final = ZCLAttributeDef(id=0x0505, type=t.uint16_t, access="rp")
        rms_voltage_min: Final = ZCLAttributeDef(id=0x0506, type=t.uint16_t, access="r")
        rms_voltage_max: Final = ZCLAttributeDef(id=0x0507, type=t.uint16_t, access="r")
        rms_current: Final = ZCLAttributeDef(id=0x0508, type=t.uint16_t, access="rp")
        rms_current_min: Final = ZCLAttributeDef(id=0x0509, type=t.uint16_t, access="r")
        rms_current_max: Final = ZCLAttributeDef(id=0x050A, type=t.uint16_t, access="r")
        active_power: Final = ZCLAttributeDef(id=0x050B, type=t.int16s, access="rp")
        active_power_min: Final = ZCLAttributeDef(id=0x050C, type=t.int16s, access="r")
        active_power_max: Final = ZCLAttributeDef(id=0x050D, type=t.int16s, access="r")
        reactive_power: Final = ZCLAttributeDef(id=0x050E, type=t.int16s, access="rp")
        apparent_power: Final = ZCLAttributeDef(id=0x050F, type=t.uint16_t, access="rp")
        power_factor: Final = ZCLAttributeDef(id=0x0510, type=t.int8s, access="r")
        average_rms_voltage_meas_period: Final = ZCLAttributeDef(
            id=0x0511, type=t.uint16_t, access="rw"
        )
        average_rms_over_voltage_counter: Final = ZCLAttributeDef(
            id=0x0512, type=t.uint16_t, access="rw"
        )
        average_rms_under_voltage_counter: Final = ZCLAttributeDef(
            id=0x0513, type=t.uint16_t, access="rw"
        )
        rms_extreme_over_voltage_period: Final = ZCLAttributeDef(
            id=0x0514, type=t.uint16_t, access="rw"
        )
        rms_extreme_under_voltage_period: Final = ZCLAttributeDef(
            id=0x0515, type=t.uint16_t, access="rw"
        )
        rms_voltage_sag_period: Final = ZCLAttributeDef(
            id=0x0516, type=t.uint16_t, access="rw"
        )
        rms_voltage_swell_period: Final = ZCLAttributeDef(
            id=0x0517, type=t.uint16_t, access="rw"
        )
        # AC Formatting
        ac_voltage_multiplier: Final = ZCLAttributeDef(
            id=0x0600, type=t.uint16_t, access="rp"
        )
        ac_voltage_divisor: Final = ZCLAttributeDef(
            id=0x0601, type=t.uint16_t, access="rp"
        )
        ac_current_multiplier: Final = ZCLAttributeDef(
            id=0x0602, type=t.uint16_t, access="rp"
        )
        ac_current_divisor: Final = ZCLAttributeDef(
            id=0x0603, type=t.uint16_t, access="rp"
        )
        ac_power_multiplier: Final = ZCLAttributeDef(
            id=0x0604, type=t.uint16_t, access="rp"
        )
        ac_power_divisor: Final = ZCLAttributeDef(
            id=0x0605, type=t.uint16_t, access="rp"
        )
        # DC Manufacturer Threshold Alarms
        dc_overload_alarms_mask: Final = ZCLAttributeDef(
            id=0x0700, type=DCOverloadAlarmMark, access="rp"
        )
        dc_voltage_overload: Final = ZCLAttributeDef(
            id=0x0701, type=t.int16s, access="rp"
        )
        dc_current_overload: Final = ZCLAttributeDef(
            id=0x0702, type=t.int16s, access="rp"
        )
        # AC Manufacturer Threshold Alarms
        ac_alarms_mask: Final = ZCLAttributeDef(
            id=0x0800, type=ACAlarmsMask, access="rw"
        )
        ac_voltage_overload: Final = ZCLAttributeDef(
            id=0x0801, type=t.int16s, access="r"
        )
        ac_current_overload: Final = ZCLAttributeDef(
            id=0x0802, type=t.int16s, access="r"
        )
        ac_active_power_overload: Final = ZCLAttributeDef(
            id=0x0803, type=t.int16s, access="r"
        )
        ac_reactive_power_overload: Final = ZCLAttributeDef(
            id=0x0804, type=t.int16s, access="r"
        )
        average_rms_over_voltage: Final = ZCLAttributeDef(
            id=0x0805, type=t.int16s, access="r"
        )
        average_rms_under_voltage: Final = ZCLAttributeDef(
            id=0x0806, type=t.int16s, access="r"
        )
        rms_extreme_over_voltage: Final = ZCLAttributeDef(
            id=0x0807, type=t.int16s, access="rw"
        )
        rms_extreme_under_voltage: Final = ZCLAttributeDef(
            id=0x0808, type=t.int16s, access="rw"
        )
        rms_voltage_sag: Final = ZCLAttributeDef(id=0x0809, type=t.int16s, access="rw")
        rms_voltage_swell: Final = ZCLAttributeDef(
            id=0x080A, type=t.int16s, access="rw"
        )
        # AC Phase B Measurements
        line_current_ph_b: Final = ZCLAttributeDef(
            id=0x0901, type=t.uint16_t, access="rp"
        )
        active_current_ph_b: Final = ZCLAttributeDef(
            id=0x0902, type=t.int16s, access="rp"
        )
        reactive_current_ph_b: Final = ZCLAttributeDef(
            id=0x0903, type=t.int16s, access="rp"
        )
        rms_voltage_ph_b: Final = ZCLAttributeDef(
            id=0x0905, type=t.uint16_t, access="rp"
        )
        rms_voltage_min_ph_b: Final = ZCLAttributeDef(
            id=0x0906, type=t.uint16_t, access="r"
        )
        rms_voltage_max_ph_b: Final = ZCLAttributeDef(
            id=0x0907, type=t.uint16_t, access="r"
        )
        rms_current_ph_b: Final = ZCLAttributeDef(
            id=0x0908, type=t.uint16_t, access="rp"
        )
        rms_current_min_ph_b: Final = ZCLAttributeDef(
            id=0x0909, type=t.uint16_t, access="r"
        )
        rms_current_max_ph_b: Final = ZCLAttributeDef(
            id=0x090A, type=t.uint16_t, access="r"
        )
        active_power_ph_b: Final = ZCLAttributeDef(
            id=0x090B, type=t.int16s, access="rp"
        )
        active_power_min_ph_b: Final = ZCLAttributeDef(
            id=0x090C, type=t.int16s, access="r"
        )
        active_power_max_ph_b: Final = ZCLAttributeDef(
            id=0x090D, type=t.int16s, access="r"
        )
        reactive_power_ph_b: Final = ZCLAttributeDef(
            id=0x090E, type=t.int16s, access="rp"
        )
        apparent_power_ph_b: Final = ZCLAttributeDef(
            id=0x090F, type=t.uint16_t, access="rp"
        )
        power_factor_ph_b: Final = ZCLAttributeDef(id=0x0910, type=t.int8s, access="r")
        average_rms_voltage_measure_period_ph_b: Final = ZCLAttributeDef(
            id=0x0911, type=t.uint16_t, access="rw"
        )
        average_rms_over_voltage_counter_ph_b: Final = ZCLAttributeDef(
            id=0x0912, type=t.uint16_t, access="rw"
        )
        average_under_voltage_counter_ph_b: Final = ZCLAttributeDef(
            id=0x0913, type=t.uint16_t, access="rw"
        )
        rms_extreme_over_voltage_period_ph_b: Final = ZCLAttributeDef(
            id=0x0914, type=t.uint16_t, access="rw"
        )
        rms_extreme_under_voltage_period_ph_b: Final = ZCLAttributeDef(
            id=0x0915, type=t.uint16_t, access="rw"
        )
        rms_voltage_sag_period_ph_b: Final = ZCLAttributeDef(
            id=0x0916, type=t.uint16_t, access="rw"
        )
        rms_voltage_swell_period_ph_b: Final = ZCLAttributeDef(
            id=0x0917, type=t.uint16_t, access="rw"
        )
        # AC Phase C Measurements
        line_current_ph_c: Final = ZCLAttributeDef(
            id=0x0A01, type=t.uint16_t, access="rp"
        )
        active_current_ph_c: Final = ZCLAttributeDef(
            id=0x0A02, type=t.int16s, access="rp"
        )
        reactive_current_ph_c: Final = ZCLAttributeDef(
            id=0x0A03, type=t.int16s, access="rp"
        )
        rms_voltage_ph_c: Final = ZCLAttributeDef(
            id=0x0A05, type=t.uint16_t, access="rp"
        )
        rms_voltage_min_ph_c: Final = ZCLAttributeDef(
            id=0x0A06, type=t.uint16_t, access="r"
        )
        rms_voltage_max_ph_c: Final = ZCLAttributeDef(
            id=0x0A07, type=t.uint16_t, access="r"
        )
        rms_current_ph_c: Final = ZCLAttributeDef(
            id=0x0A08, type=t.uint16_t, access="rp"
        )
        rms_current_min_ph_c: Final = ZCLAttributeDef(
            id=0x0A09, type=t.uint16_t, access="r"
        )
        rms_current_max_ph_c: Final = ZCLAttributeDef(
            id=0x0A0A, type=t.uint16_t, access="r"
        )
        active_power_ph_c: Final = ZCLAttributeDef(
            id=0x0A0B, type=t.int16s, access="rp"
        )
        active_power_min_ph_c: Final = ZCLAttributeDef(
            id=0x0A0C, type=t.int16s, access="r"
        )
        active_power_max_ph_c: Final = ZCLAttributeDef(
            id=0x0A0D, type=t.int16s, access="r"
        )
        reactive_power_ph_c: Final = ZCLAttributeDef(
            id=0x0A0E, type=t.int16s, access="rp"
        )
        apparent_power_ph_c: Final = ZCLAttributeDef(
            id=0x0A0F, type=t.uint16_t, access="rp"
        )
        power_factor_ph_c: Final = ZCLAttributeDef(id=0x0A10, type=t.int8s, access="r")
        average_rms_voltage_meas_period_ph_c: Final = ZCLAttributeDef(
            id=0x0A11, type=t.uint16_t, access="rw"
        )
        average_rms_over_voltage_counter_ph_c: Final = ZCLAttributeDef(
            id=0x0A12, type=t.uint16_t, access="rw"
        )
        average_under_voltage_counter_ph_c: Final = ZCLAttributeDef(
            id=0x0A13, type=t.uint16_t, access="rw"
        )
        rms_extreme_over_voltage_period_ph_c: Final = ZCLAttributeDef(
            id=0x0A14, type=t.uint16_t, access="rw"
        )
        rms_extreme_under_voltage_period_ph_c: Final = ZCLAttributeDef(
            id=0x0A15, type=t.uint16_t, access="rw"
        )
        rms_voltage_sag_period_ph_c: Final = ZCLAttributeDef(
            id=0x0A16, type=t.uint16_t, access="rw"
        )
        rms_voltage_swell_period_ph_c: Final = ZCLAttributeDef(
            id=0x0A17, type=t.uint16_t, access="rw"
        )
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR

    class ServerCommandDefs(BaseCommandDefs):
        get_profile_info: Final = ZCLCommandDef(id=0x00, schema={})
        get_measurement_profile: Final = ZCLCommandDef(id=0x01, schema={})

    class ClientCommandDefs(BaseCommandDefs):
        get_profile_info_response: Final = ZCLCommandDef(id=0x00, schema={})
        get_measurement_profile_response: Final = ZCLCommandDef(id=0x01, schema={})


class Diagnostic(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0B05
    ep_attribute: Final = "diagnostic"

    class AttributeDefs(BaseAttributeDefs):
        # Hardware Information
        number_of_resets: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="r"
        )
        persistent_memory_writes: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r"
        )
        # Stack/Network Information
        mac_rx_bcast: Final = ZCLAttributeDef(id=0x0100, type=t.uint32_t, access="r")
        mac_tx_bcast: Final = ZCLAttributeDef(id=0x0101, type=t.uint32_t, access="r")
        mac_rx_ucast: Final = ZCLAttributeDef(id=0x0102, type=t.uint32_t, access="r")
        mac_tx_ucast: Final = ZCLAttributeDef(id=0x0103, type=t.uint32_t, access="r")
        mac_tx_ucast_retry: Final = ZCLAttributeDef(
            id=0x0104, type=t.uint16_t, access="r"
        )
        mac_tx_ucast_fail: Final = ZCLAttributeDef(
            id=0x0105, type=t.uint16_t, access="r"
        )
        aps_rx_bcast: Final = ZCLAttributeDef(id=0x0106, type=t.uint16_t, access="r")
        aps_tx_bcast: Final = ZCLAttributeDef(id=0x0107, type=t.uint16_t, access="r")
        aps_rx_ucast: Final = ZCLAttributeDef(id=0x0108, type=t.uint16_t, access="r")
        aps_tx_ucast_success: Final = ZCLAttributeDef(
            id=0x0109, type=t.uint16_t, access="r"
        )
        aps_tx_ucast_retry: Final = ZCLAttributeDef(
            id=0x010A, type=t.uint16_t, access="r"
        )
        aps_tx_ucast_fail: Final = ZCLAttributeDef(
            id=0x010B, type=t.uint16_t, access="r"
        )
        route_disc_initiated: Final = ZCLAttributeDef(
            id=0x010C, type=t.uint16_t, access="r"
        )
        neighbor_added: Final = ZCLAttributeDef(id=0x010D, type=t.uint16_t, access="r")
        neighbor_removed: Final = ZCLAttributeDef(
            id=0x010E, type=t.uint16_t, access="r"
        )
        neighbor_stale: Final = ZCLAttributeDef(id=0x010F, type=t.uint16_t, access="r")
        join_indication: Final = ZCLAttributeDef(id=0x0110, type=t.uint16_t, access="r")
        child_moved: Final = ZCLAttributeDef(id=0x0111, type=t.uint16_t, access="r")
        nwk_fc_failure: Final = ZCLAttributeDef(id=0x0112, type=t.uint16_t, access="r")
        aps_fc_failure: Final = ZCLAttributeDef(id=0x0113, type=t.uint16_t, access="r")
        aps_unauthorized_key: Final = ZCLAttributeDef(
            id=0x0114, type=t.uint16_t, access="r"
        )
        nwk_decrypt_failures: Final = ZCLAttributeDef(
            id=0x0115, type=t.uint16_t, access="r"
        )
        aps_decrypt_failures: Final = ZCLAttributeDef(
            id=0x0116, type=t.uint16_t, access="r"
        )
        packet_buffer_allocate_failures: Final = ZCLAttributeDef(
            id=0x0117, type=t.uint16_t, access="r"
        )
        relayed_ucast: Final = ZCLAttributeDef(id=0x0118, type=t.uint16_t, access="r")
        phy_to_mac_queue_limit_reached: Final = ZCLAttributeDef(
            id=0x0119, type=t.uint16_t, access="r"
        )
        packet_validate_drop_count: Final = ZCLAttributeDef(
            id=0x011A, type=t.uint16_t, access="r"
        )
        average_mac_retry_per_aps_message_sent: Final = ZCLAttributeDef(
            id=0x011B, type=t.uint16_t, access="r"
        )
        last_message_lqi: Final = ZCLAttributeDef(id=0x011C, type=t.uint8_t, access="r")
        last_message_rssi: Final = ZCLAttributeDef(id=0x011D, type=t.int8s, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR



from __future__ import annotations

import dataclasses
from datetime import UTC, datetime
import enum
import typing

import attrs

from . import basic
from .struct import Struct

if typing.TYPE_CHECKING:
    from typing import Self


class Bytes(bytes):
    """A serializable type to consume all remaining bytes."""

    def serialize(self) -> Bytes:
        return self

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[Self, bytes]:
        return cls(data), b""


class BaseDataclassMixin:
    def replace(self, **kwargs: typing.Any) -> Self:
        if dataclasses.is_dataclass(self):
            assert not isinstance(self, type)  # `is_dataclass` works on types as well
            return dataclasses.replace(self, **kwargs)
        else:
            return attrs.evolve(self, **kwargs)


def _hex_string_to_bytes(hex_string: str) -> bytes:
    """Parses a hex string with optional colon delimiters and whitespace into bytes."""

    # Strips out whitespace and colons
    cleaned = "".join(hex_string.replace(":", "").split()).upper()
    return bytes.fromhex(cleaned)


class BroadcastAddress(basic.enum16):
    ALL_DEVICES = 0xFFFF
    RESERVED_FFFE = 0xFFFE
    RX_ON_WHEN_IDLE = 0xFFFD
    ALL_ROUTERS_AND_COORDINATOR = 0xFFFC
    LOW_POWER_ROUTER = 0xFFFB
    RESERVED_FFFA = 0xFFFA
    RESERVED_FFF9 = 0xFFF9
    RESERVED_FFF8 = 0xFFF8


class EUI64(basic.FixedList, item_type=basic.uint8_t, length=8):
    # EUI 64-bit ID (an IEEE address).
    def __repr__(self) -> str:
        return ":".join(f"{i:02x}" for i in self[::-1])

    def __hash__(self) -> int:  # type: ignore[override]
        return hash(repr(self))

    @classmethod
    def convert(cls, ieee: str) -> EUI64:
        if ieee is None:
            return None
        ieee = [basic.uint8_t(p) for p in _hex_string_to_bytes(ieee)[::-1]]
        assert len(ieee) == cls._length
        return cls(ieee)


EUI64.UNKNOWN = EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF")


class KeyData(basic.FixedList, item_type=basic.uint8_t, length=16):
    def __repr__(self) -> str:
        return ":".join(f"{i:02x}" for i in self)

    @classmethod
    def convert(cls, key: str) -> KeyData:
        key = [basic.uint8_t(p) for p in _hex_string_to_bytes(key)]
        assert len(key) == cls._length
        return cls(key)


KeyData.UNKNOWN = KeyData.convert("FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF")


class Bool(basic.enum8):
    false = 0
    true = 1


class AttributeId(basic.uint16_t, repr="hex"):
    pass


class BACNetOid(basic.uint32_t):
    pass


class Channels(basic.bitmap32):
    """Zigbee Channels."""

    NO_CHANNELS = 0x00000000
    ALL_CHANNELS = 0x07FFF800
    CHANNEL_11 = 0x00000800
    CHANNEL_12 = 0x00001000
    CHANNEL_13 = 0x00002000
    CHANNEL_14 = 0x00004000
    CHANNEL_15 = 0x00008000
    CHANNEL_16 = 0x00010000
    CHANNEL_17 = 0x00020000
    CHANNEL_18 = 0x00040000
    CHANNEL_19 = 0x00080000
    CHANNEL_20 = 0x00100000
    CHANNEL_21 = 0x00200000
    CHANNEL_22 = 0x00400000
    CHANNEL_23 = 0x00800000
    CHANNEL_24 = 0x01000000
    CHANNEL_25 = 0x02000000
    CHANNEL_26 = 0x04000000

    @classmethod
    def from_channel_list(cls: Channels, channels: typing.Iterable[int]) -> Channels:
        mask = cls.NO_CHANNELS

        for channel in channels:
            if not 11 <= channel <= 26:
                raise ValueError(
                    f"Invalid channel number {channel}. Must be between 11 and 26."
                )

            mask |= cls[f"CHANNEL_{channel}"]  # type: ignore[index]

        return mask

    def __iter__(self):
        cls = type(self)

        channels = [c for c in range(11, 26 + 1) if self & cls[f"CHANNEL_{c}"]]

        if self != cls.from_channel_list(channels):
            raise ValueError(f"Channels bitmap has unexpected members: {self}")

        return iter(channels)


class ClusterId(basic.uint16_t):
    pass


class Date(Struct):
    years_since_1900: basic.uint8_t
    month: basic.uint8_t
    day: basic.uint8_t
    day_of_week: basic.uint8_t

    @property
    def year(self):
        if self.years_since_1900 is None:
            return None

        return 1900 + self.years_since_1900

    @year.setter
    def year(self, years):
        assert 1900 <= years <= 2155
        self.years_since_1900 = years - 1900


class NWK(basic.uint16_t, repr="hex"):
    @classmethod
    def convert(cls, data: str) -> NWK:
        assert 4 * len(data) == cls._bits
        return cls.deserialize(bytes.fromhex(data)[::-1])[0]


class PanId(NWK):
    pass


class ExtendedPanId(EUI64):
    pass


class Group(basic.uint16_t, repr="hex"):
    pass


class NoData:
    @classmethod
    def deserialize(cls, data):
        return cls(), data

    def serialize(self):
        return b""


class TimeOfDay(Struct):
    hours: basic.uint8_t
    minutes: basic.uint8_t
    seconds: basic.uint8_t
    hundredths: basic.uint8_t


class _Time(basic.uint32_t):
    pass


class UTCTime(_Time):
    pass


class StandardTime(_Time):
    """Adjusted for TimeZone but not for daylight saving."""


class LocalTime(_Time):
    """Standard time adjusted for daylight saving."""


class Relays(basic.LVList, item_type=NWK, length_type=basic.uint8_t):
    """Relay list for static routing."""


class APSStatus(basic.enum8):
    # A request has been executed successfully
    APS_SUCCESS = 0x00

    # A transmit request failed since the ASDU is too large and fragmentation
    # is not supported
    APS_ASDU_TOO_LONG = 0xA0

    # A received fragmented frame could not be defragmented at the current time
    APS_DEFRAG_DEFERRED = 0xA1

    # A received fragmented frame could not be defragmented since the device
    # does not support fragmentation
    APS_DEFRAG_UNSUPPORTED = 0xA2

    # A parameter value was out of range
    APS_ILLEGAL_REQUEST = 0xA3

    # An APSME-UNBIND.request failed due to the requested binding link not
    # existing in the binding table
    APS_INVALID_BINDING = 0xA4

    # An APSME-REMOVE-GROUP.request has been issued with a group identifier
    # that does not appear in the group table
    APS_INVALID_GROUP = 0xA5

    # A parameter value was invalid or out of range
    APS_INVALID_PARAMETER = 0xA6

    # An APSDE-DATA.request requesting acknowledged transmission failed due to
    # no acknowledgement being received
    APS_NO_ACK = 0xA7

    # An APSDE-DATA.request with a destination addressing mode set to 0x00
    # failed due to there being no devices bound to this device
    APS_NO_BOUND_DEVICE = 0xA8

    # An APSDE-DATA.request with a destination addressing mode set to 0x03
    # failed due to no corresponding short address found in the address map
    # table
    APS_NO_SHORT_ADDRESS = 0xA9

    # An APSDE-DATA.request with a destination addressing mode set to 0x00
    # failed due to a binding table not being supported on the device
    APS_NOT_SUPPORTED = 0xAA

    # An ASDU was received that was secured using a link key
    APS_SECURED_LINK_KEY = 0xAB

    # An ASDU was received that was secured using a network key
    APS_SECURED_NWK_KEY = 0xAC

    #  An APSDE-DATA.request requesting security has resulted in an error
    #  during the corresponding security processing
    APS_SECURITY_FAIL = 0xAD

    # An APSME-BIND.request or APSME.ADDGROUP.request issued when the binding
    # or group tables, respectively, were full
    APS_TABLE_FULL = 0xAE

    # An ASDU was received without any security
    APS_UNSECURED = 0xAF

    # An APSME-GET.request or APSMESET.request has been issued with an unknown
    # attribute identifier
    APS_UNSUPPORTED_ATTRIBUTE = 0xB0

    @classmethod
    def _missing_(cls, value):
        chained = NWKStatus(value)
        status = cls._member_type_.__new__(cls, chained.value)
        status._name_ = chained.name
        status._value_ = value
        return status


class MACStatus(basic.enum8):
    # Operation was successful
    MAC_SUCCESS = 0x00

    # Association Status field
    MAC_PAN_AT_CAPACITY = 0x01
    MAC_PAN_ACCESS_DENIED = 0x02

    # The frame counter purportedly applied by the originator of the received
    # frame is invalid
    MAC_COUNTER_ERROR = 0xDB

    # The key purportedly applied by the originator of the received frame is
    # not allowed to be used with that frame type according to the key usage
    # policy of the recipient
    MAC_IMPROPER_KEY_TYPE = 0xDC

    # The security level purportedly applied # by the originator of the
    # received frame does not meet the minimum security level
    # required/expected by the recipient for that frame type
    MAC_IMPROPER_SECURITY_LEVEL = 0xDD

    # The received frame was purportedly secured using security based on IEEE
    # Std 802.15.4-2003, and such security is not supported by this standard
    MAC_UNSUPPORTED_LEGACY = 0xDE

    # The security purportedly applied by the originator of the received frame
    # is not supported
    MAC_UNSUPPORTED_SECURITY = 0xDF

    # The beacon was lost following a synchronization request
    MAC_BEACON_LOSS = 0xE0

    # A transmission could not take place due to activity on the channel, i.e.
    # the CSMA-CA mechanism has failed
    MAC_CHANNEL_ACCESS_FAILURE = 0xE1

    # The GTS request has been denied by the PAN coordinator
    MAC_DENIED = 0xE2

    # The attempt to disable the transceiver has failed
    MAC_DISABLE_TRX_FAILURE = 0xE3

    # Cryptographic processing of the received secured frame failed
    MAC_SECURITY_ERROR = 0xE4

    # Either a frame resulting from processing has a length that is greater
    # than aMaxPHYPacketSize or a requested transaction is too large to fit in
    # the CAP or GTS
    MAC_FRAME_TOO_LONG = 0xE5

    # The requested GTS transmission failed because the specified GTS either
    # did not have a transmit GTS direction or was not defined
    MAC_INVALID_GTS = 0xE6

    # A request to purge an MSDU from the transaction queue was made using an
    # MSDU handle that was not found in the transaction table
    MAC_INVALID_HANDLE = 0xE7

    # A parameter in the primitive is either not supported or is out of the
    # valid range
    MAC_INVALID_PARAMETER = 0xE8

    # No acknowledgment was received after macMaxFrameRetries
    MAC_NO_ACK = 0xE9

    # A scan operation failed to find any network beacons
    MAC_NO_BEACON = 0xEA

    # No response data was available following a request
    MAC_NO_DATA = 0xEB

    # The operation failed because a 16-bit short address was not allocated
    MAC_NO_SHORT_ADDRESS = 0xEC

    # A receiver enable request was unsuccessful because it could not be
    # completed within the CAP. @note The enumeration description is not used
    # in this standard, and it is included only to meet the backwards
    # compatibility requirements for IEEE Std 802.15.4-2003
    MAC_OUT_OF_CAP = 0xED

    # A PAN identifier conflict has been detected and communicated to the PAN
    # coordinator
    MAC_PAN_ID_CONFLICT = 0xEE

    # A coordinator realignment command has been received
    MAC_REALIGNMENT = 0xEF

    # The transaction has expired and its information was discarded
    MAC_TRANSACTION_EXPIRED = 0xF0

    # There is no capacity to store the transaction
    MAC_TRANSACTION_OVERFLOW = 0xF1

    # The transceiver was in the transmitter enabled state when the receiver
    # was requested to be enabled. @note The enumeration description is not
    # used in this standard, and it is included only to meet the backwards
    # compatibility requirements for IEEE Std 802.15.4-2003
    MAC_TX_ACTIVE = 0xF2

    # The key purportedly used by the originator of the received frame is not
    # available or, if available, the originating device is not known or is
    # blacklisted with that particular key
    MAC_UNAVAILABLE_KEY = 0xF3

    # A SET/GET request was issued with the identifier of a PIB attribute that
    # is not supported
    MAC_UNSUPPORTED_ATTRIBUTE = 0xF4

    # A request to send data was unsuccessful because neither the source
    # address parameters nor the destination address parameters were present
    MAC_INVALID_ADDRESS = 0xF5

    # A receiver enable request was unsuccessful because it specified a number
    # of symbols that was longer than the beacon interval
    MAC_ON_TIME_TOO_LONG = 0xF6

    # A receiver enable request was unsuccessful because it could not be
    # completed within the current superframe and was not permitted to be
    # deferred until the next superframe
    MAC_PAST_TIME = 0xF7

    # The device was instructed to start sending beacons based on the timing
    # of the beacon transmissions of its coordinator, but the device is not
    # currently tracking the beacon of its coordinator
    MAC_TRACKING_OFF = 0xF8

    # An attempt to write to a MAC PIB attribute that is in a table failed
    # because the specified table index was out of range
    MAC_INVALID_INDEX = 0xF9

    # A scan operation terminated prematurely because the number of PAN
    # descriptors stored reached an implementation specified maximum
    MAC_LIMIT_REACHED = 0xFA

    # A SET/GET request was issued with the identifier of an attribute that is
    # read only
    MAC_READ_ONLY = 0xFB

    # A request to perform a scan operation failed because the MLME was in the
    # process of performing a previously initiated scan operation
    MAC_SCAN_IN_PROGRESS = 0xFC

    # The device was instructed to start sending beacons based on the timing
    # of the beacon transmissions of its coordinator, but the instructed start
    # time overlapped the transmission time of the beacon of its coordinator
    MAC_SUPERFRAME_OVERLAP = 0xFD


class NWKStatus(basic.enum8):
    # A request has been executed successfully
    NWK_SUCCESS = 0x00

    # An invalid or out-of-range parameter has been passed to a primitive from
    # the next higher layer
    NWK_INVALID_PARAMETER = 0xC1

    # The next higher layer has issued a request that is invalid or cannot be
    # executed given the current state of the NWK layer
    NWK_INVALID_REQUEST = 0xC2

    # An NLME-JOIN.request has been disallowed
    NWK_NOT_PERMITTED = 0xC3

    # An NLME-NETWORK-FORMATION.request has failed to start a network
    NWK_STARTUP_FAILURE = 0xC4

    # A device with the address supplied to the NLMEDIRECT-JOIN.request is
    # already present in the neighbor table of the device on which the
    # NLMEDIRECT-JOIN.request was issued
    NWK_ALREADY_PRESENT = 0xC5

    # Used to indicate that an NLME-SYNC.request has failed at the MAC layer
    NWK_SYNC_FAILURE = 0xC6

    # An NLME-JOIN-DIRECTLY.request has failed because there is no more room
    # in the neighbor table
    NWK_NEIGHBOR_TABLE_FULL = 0xC7

    # An NLME-LEAVE.request has failed because the device addressed in the
    # parameter list is not in the neighbor table of the issuing device
    NWK_UNKNOWN_DEVICE = 0xC8

    # An NLME-GET.request or NLME-SET.request has been issued with an unknown
    # attribute identifier
    NWK_UNSUPPORTED_ATTRIBUTE = 0xC9

    # An NLME-JOIN.request has been issued in an environment where no networks
    # are detectable
    NWK_NO_NETWORKS = 0xCA

    NWK_RESERVED_0xCB = 0xCB

    # Security processing has been attempted on an outgoing frame, and has
    # failed because the frame counter has reached its maximum value
    NWK_NWK_MAX_FRM_COUNTER = 0xCC

    # Security processing has been attempted on an outgoing frame, and has
    # failed because no key was available with which to process it
    NWK_NO_KEY = 0xCD

    # Security processing has been attempted on an outgoing frame, and has
    # failed because the security engine produced erroneous output
    NWK_BAD_CCM_OUTPUT = 0xCE

    NWK_RESERVED_0xCF = 0xCF

    # An attempt to discover a route has failed due to a reason other than a
    # lack of routing capacity
    NWK_ROUTE_DISCOVERY_FAILED = 0xD0

    # An NLDE-DATA.request has failed due to a routing failure on the sending
    # device or an NLMEROUTE-DISCOVERY.request has failed due to the cause
    # cited in the accompanying NetworkStatusCode
    NWK_ROUTE_ERROR = 0xD1

    # An attempt to send a broadcast frame or member mode multicast has failed
    # due to the fact that there is no room in the BTT
    NWK_BT_TABLE_FULL = 0xD2

    # An NLDE-DATA.request has failed due to insufficient buffering available.
    # A non-member mode multicast frame was discarded pending route discovery
    NWK_FRAME_NOT_BUFFERED = 0xD3

    @classmethod
    def _missing_(cls, value):
        chained = MACStatus(value)
        status = cls._member_type_.__new__(cls, chained.value)
        status._name_ = chained.name
        status._value_ = value
        return status


class AddrMode(basic.enum8):
    """Addressing mode."""

    Group = 0x01
    NWK = 0x02
    IEEE = 0x03
    Broadcast = 0x0F


class Addressing:
    """Deprecated, only present for backwards compatibility."""

    Group = AddrMode
    NWK = AddrMode
    IEEE = AddrMode
    Broadcast = AddrMode


@dataclasses.dataclass
class AddrModeAddress(BaseDataclassMixin):
    """Address mode and address."""

    addr_mode: AddrMode
    address: NWK | Group | EUI64 | BroadcastAddress | None

    def __post_init__(self) -> None:
        if self.addr_mode is not None and self.address is not None:
            self.address = {
                AddrMode.Group: Group,
                AddrMode.NWK: NWK,
                AddrMode.IEEE: EUI64,
                AddrMode.Broadcast: BroadcastAddress,
            }[self.addr_mode](self.address)

    def __hash__(self) -> int:
        return hash((self.addr_mode, self.address))


class TransmitOptions(enum.Flag):
    NONE = 0

    ACK = 1 << 0
    APS_Encryption = 1 << 1
    FORCE_ROUTE_DISCOVERY = 1 << 2


class PacketPriority(enum.IntEnum):
    """Packet priority"""

    CRITICAL = 2
    HIGH = 1
    NORMAL = 0
    LOW = -1


@dataclasses.dataclass
class ZigbeePacket(BaseDataclassMixin):
    """Container for the information in an incoming or outgoing ZDO or ZCL packet.

    The radio library is expected to fill this object in with all received data and pass
    it to zigpy for every type of packet.
    """

    timestamp: datetime = dataclasses.field(
        compare=False, default_factory=lambda: datetime.now(UTC)
    )

    # Higher priority will try to be sent before lower
    priority: int | None = dataclasses.field(default=None)

    # Set to `None` when the packet is outgoing
    src: AddrModeAddress | None = dataclasses.field(default=None)
    src_ep: basic.uint8_t | None = dataclasses.field(default=None)

    # Set to `None` when the packet is incoming
    dst: AddrModeAddress | None = dataclasses.field(default=None)
    dst_ep: basic.uint8_t | None = dataclasses.field(default=None)

    # If the radio supports it, a source route for the packet
    source_route: list[NWK] | None = dataclasses.field(default=None)
    extended_timeout: bool = dataclasses.field(default=False)

    tsn: basic.uint8_t = dataclasses.field(default=0x00)
    profile_id: basic.uint16_t = dataclasses.field(default=0x0000)
    cluster_id: basic.uint16_t = dataclasses.field(default=0x0000)

    # Any serializable object
    data: basic.SerializableBytes = dataclasses.field(
        default_factory=basic.SerializableBytes
    )

    # Options for outgoing packets
    tx_options: TransmitOptions = dataclasses.field(default=TransmitOptions.NONE)
    radius: basic.uint8_t = dataclasses.field(default=0)
    non_member_radius: basic.uint8_t = dataclasses.field(default=0)

    # Options for incoming packets
    lqi: basic.uint8_t | None = dataclasses.field(default=None)
    rssi: basic.int8s | None = dataclasses.field(default=None)

    def __hash__(self) -> int:
        return hash(
            (
                self.timestamp,
                self.src,
                self.src_ep,
                self.dst,
                self.dst_ep,
                self.source_route,
                self.extended_timeout,
                self.tsn,
                self.profile_id,
                self.cluster_id,
                self.data,
                self.tx_options,
                self.radius,
                self.non_member_radius,
                self.lqi,
                self.rssi,
                self.priority,
            )
        )


@dataclasses.dataclass(frozen=True)
class NetworkBeacon(BaseDataclassMixin):
    pan_id: PanId
    extended_pan_id: EUI64
    channel: basic.uint8_t
    permit_joining: bool
    stack_profile: basic.uint8_t
    nwk_update_id: basic.uint8_t
    lqi: basic.uint8_t

    # Migrate to kwarg-only once we drop 3.9
    src: NWK | None = None
    rssi: basic.int8s | None = None
    depth: basic.uint8_t | None = None
    router_capacity: bool | None = None
    device_capacity: bool | None = None
    protocol_version: basic.uint8_t | None = None


@dataclasses.dataclass(frozen=True)
class CapturedPacket(BaseDataclassMixin):
    timestamp: datetime
    rssi: float
    lqi: basic.uint8_t
    channel: basic.uint8_t
    data: bytes

    def compute_fcs(self) -> bytes:
        crc = 0x0000

        for c in self.data:
            q = (crc ^ c) & 15  # Do low-order 4 bits
            crc = (crc // 16) ^ (q * 0x1081)

            q = (crc ^ (c // 16)) & 15  # And high 4 bits
            crc = (crc // 16) ^ (q * 0x1081)

        return crc.to_bytes(2, "little")



"""Common fixtures."""

from __future__ import annotations

import asyncio
import contextlib
from contextlib import contextmanager
import copy
import logging
import sqlite3
import threading
import typing
from unittest.mock import Mock, patch

import aiosqlite
import pytest

import zigpy.application
from zigpy.config import (
    CONF_DATABASE,
    CONF_DEVICE,
    CONF_DEVICE_PATH,
    CONF_OTA,
    CONF_OTA_ENABLED,
)
import zigpy.state as app_state
import zigpy.types as t
from zigpy.typing import UNDEFINED, UndefinedType
from zigpy.zcl import Cluster, foundation
import zigpy.zdo.types as zdo_t

from .async_mock import AsyncMock, MagicMock

if typing.TYPE_CHECKING:
    import zigpy.device

_LOGGER = logging.getLogger(__name__)


@pytest.fixture
def auto_kill_aiosqlite():
    """Aiosqlite's background thread does not let pytest exit when a failure occurs."""
    yield

    for thread in threading.enumerate():
        if not isinstance(thread, aiosqlite.core.Connection):
            continue

        try:
            conn = thread._conn
        except ValueError:
            pass
        else:
            with contextlib.suppress(sqlite3.ProgrammingError):
                conn.close()

        thread._stop_running()
        thread.join(timeout=1)


NCP_IEEE = t.EUI64.convert("aa:11:22:bb:33:44:be:ef")


class FailOnBadFormattingHandler(logging.Handler):
    def emit(self, record):
        try:
            record.msg % record.args
        except Exception as e:  # noqa: BLE001
            pytest.fail(
                f"Failed to format log message {record.msg!r} with {record.args!r}: {e}"
            )


@pytest.fixture(autouse=True)
def raise_on_bad_log_formatting():
    handler = FailOnBadFormattingHandler()

    root = logging.getLogger()
    root.addHandler(handler)
    root.setLevel(logging.DEBUG)

    try:
        yield
    finally:
        root.removeHandler(handler)


class BaseApp(zigpy.application.ControllerApplication):
    async def send_packet(self, packet):
        async with self._limit_concurrency(priority=packet.priority):
            return await self._send_packet(packet)

    async def _send_packet(self, packet):
        await asyncio.sleep(0)

    async def connect(self):
        pass

    async def disconnect(self):
        pass

    async def start_network(self):
        dev = add_initialized_device(
            app=self, nwk=self.state.node_info.nwk, ieee=self.state.node_info.ieee
        )
        dev.model = "Coordinator Model"
        dev.manufacturer = "Coordinator Manufacturer"

        dev.zdo.Mgmt_NWK_Update_req = AsyncMock(
            return_value=[
                zdo_t.Status.SUCCESS,
                t.Channels.ALL_CHANNELS,
                0,
                0,
                [80] * 16,
            ]
        )

    async def force_remove(self, dev):
        pass

    async def add_endpoint(self, descriptor):
        pass

    async def permit_ncp(self, time_s=60):
        pass

    async def permit_with_link_key(self, node, link_key, time_s=60):
        pass

    async def reset_network_info(self):
        pass

    async def write_network_info(self, *, network_info, node_info):
        pass

    async def load_network_info(self, *, load_devices=False):
        self.state.network_info.channel = 15


class FeaturelessApp(BaseApp):
    """Zigpy application implementation that supports no optional features."""


class App(BaseApp):
    """Zigpy application implementation that supports all optional features."""

    async def _network_scan(self, channels, duration_exp):
        if False:
            yield

    async def _packet_capture(self, channel):
        if False:
            yield

    async def _packet_capture_change_channel(self, channel):
        pass

    async def _set_tx_power(self, tx_power: float) -> float:
        return round(tx_power)

    async def _get_recommended_tx_power(self, country: str) -> float:
        if country == "US":
            return 8
        else:
            return 10

    async def _get_maximum_tx_power(self, country: str) -> float:
        if country == "US":
            return 8
        else:
            return 10


def recursive_dict_merge(
    obj: dict[str, typing.Any], updates: dict[str, typing.Any]
) -> dict[str, typing.Any]:
    result = copy.deepcopy(obj)

    for key, update in updates.items():
        if isinstance(update, dict) and key in result:
            result[key] = recursive_dict_merge(result[key], update)
        else:
            result[key] = update

    return result


def make_app(
    config_updates: dict[str, typing.Any],
    app_base: type[zigpy.application.ControllerApplication] = App,
) -> zigpy.application.ControllerApplication:
    config = recursive_dict_merge(
        {
            CONF_DATABASE: None,
            CONF_DEVICE: {CONF_DEVICE_PATH: "/dev/null"},
            CONF_OTA: {
                CONF_OTA_ENABLED: False,
            },
        },
        config_updates,
    )

    app = app_base(config)
    app.state.node_info = app_state.NodeInfo(
        nwk=t.NWK(0x0000), ieee=NCP_IEEE, logical_type=zdo_t.LogicalType.Coordinator
    )

    app.device_initialized = Mock(wraps=app.device_initialized)
    app.listener_event = Mock(wraps=app.listener_event)
    app.get_sequence = MagicMock(wraps=app.get_sequence, return_value=123)
    app.send_packet = AsyncMock(wraps=app.send_packet)
    app.write_network_info = AsyncMock(wraps=app.write_network_info)

    return app


@pytest.fixture
def app():
    """ControllerApplication Mock."""
    return make_app({})


@pytest.fixture
def app_mock():
    """ControllerApplication Mock."""
    return make_app({})


def make_ieee(start=0):
    return t.EUI64(map(t.uint8_t, range(start, start + 8)))


def make_node_desc(
    *,
    logical_type: zdo_t.LogicalType = zdo_t.LogicalType.Router,
    manufacturer_code: int = 4174,
) -> zdo_t.NodeDescriptor:
    return zdo_t.NodeDescriptor(
        logical_type=logical_type,
        complex_descriptor_available=0,
        user_descriptor_available=0,
        reserved=0,
        aps_flags=0,
        frequency_band=zdo_t.NodeDescriptor.FrequencyBand.Freq2400MHz,
        mac_capability_flags=zdo_t.NodeDescriptor.MACCapabilityFlags.AllocateAddress,
        manufacturer_code=manufacturer_code,
        maximum_buffer_size=82,
        maximum_incoming_transfer_size=82,
        server_mask=0,
        maximum_outgoing_transfer_size=82,
        descriptor_capability_field=zdo_t.NodeDescriptor.DescriptorCapability.NONE,
    )


def add_initialized_device(app, nwk, ieee):
    dev = app.add_device(nwk=nwk, ieee=ieee)
    dev.node_desc = make_node_desc(logical_type=zdo_t.LogicalType.Router)

    ep = dev.add_endpoint(1)
    ep.status = zigpy.endpoint.Status.ZDO_INIT
    ep.profile_id = 260
    ep.device_type = zigpy.profiles.zha.DeviceType.PUMP

    return dev


@pytest.fixture
def make_initialized_device():
    count = 1

    def inner(app):
        nonlocal count

        dev = add_initialized_device(app, nwk=0x1000 + count, ieee=make_ieee(count))
        count += 1

        return dev

    return inner


def make_neighbor(
    *,
    ieee: t.EUI64,
    nwk: t.NWK,
    device_type: zdo_t.Neighbor.DeviceType = zdo_t.Neighbor.DeviceType.Router,
    rx_on_when_idle=True,
    relationship: zdo_t.Neighbor.Relationship = zdo_t.Neighbor.Relationship.Child,
) -> zdo_t.Neighbor:
    return zdo_t.Neighbor(
        extended_pan_id=make_ieee(start=0),
        ieee=ieee,
        nwk=nwk,
        device_type=device_type,
        rx_on_when_idle=int(rx_on_when_idle),
        relationship=relationship,
        reserved1=0,
        permit_joining=0,
        reserved2=0,
        depth=15,
        lqi=250,
    )


def make_neighbor_from_device(
    device: zigpy.device.Device,
    *,
    relationship: zdo_t.Neighbor.Relationship = zdo_t.Neighbor.Relationship.Child,
):
    assert device.node_desc is not None
    return make_neighbor(
        ieee=device.ieee,
        nwk=device.nwk,
        device_type=zdo_t.Neighbor.DeviceType(int(device.node_desc.logical_type)),
        rx_on_when_idle=device.node_desc.is_receiver_on_when_idle,
        relationship=relationship,
    )


def make_route(
    *,
    dest_nwk: t.NWK,
    next_hop: t.NWK,
    status: zdo_t.RouteStatus = zdo_t.RouteStatus.Active,
) -> zdo_t.Route:
    return zdo_t.Route(
        DstNWK=dest_nwk,
        RouteStatus=status,
        MemoryConstrained=0,
        ManyToOne=0,
        RouteRecordRequired=0,
        Reserved=0,
        NextHop=next_hop,
    )


# Taken from Home Assistant's `conftest.py`
@pytest.fixture(autouse=True)
def verify_cleanup() -> typing.Generator[None, None, None]:
    """Verify that the test has cleaned up resources correctly."""

    try:
        event_loop = asyncio.get_running_loop()
    except RuntimeError:
        yield
        return

    threads_before = frozenset(threading.enumerate())
    tasks_before = asyncio.all_tasks(event_loop)
    yield

    event_loop.run_until_complete(event_loop.shutdown_default_executor())

    # Warn and clean-up lingering tasks and timers
    # before moving on to the next test.
    tasks = asyncio.all_tasks(event_loop) - tasks_before
    for task in tasks:
        _LOGGER.warning("Linger task after test %r", task)
        task.cancel()
    if tasks:
        event_loop.run_until_complete(asyncio.wait(tasks))

    for handle in event_loop._scheduled:  # type: ignore[attr-defined]
        if not handle.cancelled():
            _LOGGER.warning("Lingering timer after test %r", handle)
            handle.cancel()

    # Verify no threads were left behind.
    threads = frozenset(threading.enumerate()) - threads_before
    for thread in threads:
        if isinstance(thread, threading._DummyThread):
            continue

        # Kill lingering aiosqlite threads so pytest doesn't hang
        if isinstance(thread, aiosqlite.Connection):
            _LOGGER.warning("Stopping lingering aiosqlite thread %r", thread)
            thread._stop_running()
            thread.join(timeout=1)

        pytest.fail(f"Lingering thread after test: {thread!r}")


@contextmanager
def mock_attribute_reads(
    cluster: Cluster,
    mock_attributes: dict[str | int | foundation.ZCLAttributeDef, typing.Any],
) -> typing.Generator[
    tuple[AsyncMock, dict[str | int | foundation.ZCLAttributeDef, AsyncMock]],
    None,
    None,
]:
    """Mock attribute reads on a cluster."""
    mock_reads = {}

    for key, value in mock_attributes.items():
        if not callable(value):
            value = Mock(return_value=value)

        mock_reads[cluster.find_attribute(key)] = value

    async def read_attributes_raw(
        attributes,
        *args,
        manufacturer: int | UndefinedType | None = UNDEFINED,
        **kwargs,
    ):
        status_records = []

        for attrid in attributes:
            record = foundation.ReadAttributeRecord(attrid=attrid)
            attr_def = cluster.find_attribute(attrid, manufacturer_code=manufacturer)

            if attr_def in mock_reads:
                value = mock_reads[attr_def]()

                if value is None:
                    # Omit the record from the response entirely
                    continue
                elif isinstance(value, foundation.Status):
                    record.status = value
                else:
                    record.status = foundation.Status.SUCCESS
                    record.value = foundation.TypeValue(
                        type=attr_def.zcl_type, value=value
                    )
            else:
                record.status = foundation.Status.UNSUPPORTED_ATTRIBUTE

            status_records.append(record)

        return foundation.GENERAL_COMMANDS[
            foundation.GeneralCommand.Read_Attributes_rsp
        ].schema(status_records=status_records)

    with patch.object(
        cluster, "read_attributes_raw", autospec=True, side_effect=read_attributes_raw
    ) as mock_read:
        yield mock_read, mock_attributes


@contextmanager
def mock_attribute_writes(
    cluster: Cluster,
    mock_attributes: dict[str | int | foundation.ZCLAttributeDef, typing.Any],
) -> typing.Generator[
    tuple[AsyncMock, dict[str | int | foundation.ZCLAttributeDef, AsyncMock]],
    None,
    None,
]:
    """Mock attribute writes on a cluster."""
    mock_writes = {}

    for key, value in mock_attributes.items():
        if not callable(value):
            value = Mock(return_value=value)

        mock_writes[cluster.find_attribute(key)] = value

    async def write_attributes_raw(
        attributes,
        *args,
        manufacturer: int | UndefinedType | None = UNDEFINED,
        **kwargs,
    ):
        records = []

        for attr in attributes:
            attr_def = cluster.find_attribute(
                attr.attrid, manufacturer_code=manufacturer
            )
            record = foundation.WriteAttributesStatusRecord(attrid=attr_def.id)

            if attr_def in mock_writes:
                value = mock_writes[attr_def](attr.value)

                if isinstance(value, foundation.Status):
                    record.status = value
                else:
                    record.status = foundation.Status.SUCCESS
            else:
                record.status = foundation.Status.UNSUPPORTED_ATTRIBUTE

            records.append(record)

        return foundation.GENERAL_COMMANDS[
            foundation.GeneralCommand.Write_Attributes_rsp
        ].schema(status_records=foundation.WriteAttributesResponse(records))

    with patch.object(
        cluster, "_write_attributes", autospec=True, side_effect=write_attributes_raw
    ) as mock_write:
        yield mock_write, mock_attributes


async def mock_attribute_report(
    cluster: Cluster,
    attributes: dict[str | int | foundation.ZCLAttributeDef, typing.Any],
    *,
    tsn: int | None = None,
) -> None:
    """Mock attribute reports on a cluster."""
    reports = []
    manufacturer_codes: set[int | None] = set()

    for attr, value in attributes.items():
        if isinstance(attr, int):
            # Raw attribute ID for unknown attributes
            manufacturer_codes.add(None)
            reports.append(
                foundation.Attribute(
                    attrid=attr,
                    value=foundation.TypeValue(
                        type=foundation.DataType.from_python_type(type(value)).type_id,
                        value=value,
                    ),
                )
            )
        else:
            attr_def = cluster.find_attribute(attr)
            manufacturer_codes.add(cluster._get_effective_manufacturer_code(attr_def))

            reports.append(
                foundation.Attribute(
                    attrid=attr_def.id,
                    value=foundation.TypeValue(type=attr_def.zcl_type, value=value),
                )
            )

    if len(manufacturer_codes) != 1:
        raise ValueError(
            f"All attributes must have the same manufacturer code, got {manufacturer_codes}"
        )

    if tsn is None:
        tsn = cluster.endpoint.device.get_sequence()

    manufacturer: int | None = manufacturer_codes.pop()

    frame_control = foundation.FrameControl(
        frame_type=foundation.FrameType.GLOBAL_COMMAND,
        is_manufacturer_specific=(manufacturer is not None),
        direction=(
            foundation.Direction.Client_to_Server
            if cluster.is_client
            else foundation.Direction.Server_to_Client
        ),
        disable_default_response=False,
        reserved=0b000,
    )

    hdr = foundation.ZCLHeader(
        frame_control=frame_control,
        manufacturer=manufacturer,
        tsn=tsn,
        command_id=foundation.GeneralCommand.Report_Attributes,
    )

    command = foundation.GENERAL_COMMANDS[
        foundation.GeneralCommand.Report_Attributes
    ].schema(attribute_reports=reports)

    cluster.handle_cluster_general_request(hdr, command)



"""Measurement & Sensing Functional Domain"""

from __future__ import annotations

from typing import Final

import zigpy.types as t
from zigpy.zcl import Cluster, foundation
from zigpy.zcl.foundation import BaseAttributeDefs, ZCLAttributeDef


class LightSensorType(t.enum8):
    Photodiode = 0x00
    CMOS = 0x01
    Unknown = 0xFF


class IlluminanceMeasurement(Cluster):
    LightSensorType: Final = LightSensorType

    cluster_id: Final[t.uint16_t] = 0x0400
    name: Final = "Illuminance Measurement"
    ep_attribute: Final = "illuminance"

    class AttributeDefs(BaseAttributeDefs):
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        light_sensor_type: Final = ZCLAttributeDef(
            id=0x0004, type=LightSensorType, access="r"
        )
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class LevelStatus(t.enum8):
    Illuminance_On_Target = 0x00
    Illuminance_Below_Target = 0x01
    Illuminance_Above_Target = 0x02


class IlluminanceLevelSensing(Cluster):
    LevelStatus: Final = LevelStatus
    LightSensorType: Final = LightSensorType

    cluster_id: Final[t.uint16_t] = 0x0401
    name: Final = "Illuminance Level Sensing"
    ep_attribute: Final = "illuminance_level"

    class AttributeDefs(BaseAttributeDefs):
        level_status: Final = ZCLAttributeDef(
            id=0x0000, type=LevelStatus, access="r", mandatory=True
        )
        light_sensor_type: Final = ZCLAttributeDef(
            id=0x0001, type=LightSensorType, access="r"
        )
        # Illuminance Level Sensing Settings
        illuminance_target_level: Final = ZCLAttributeDef(
            id=0x0010, type=t.uint16_t, access="rw", mandatory=True
        )
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class TemperatureMeasurement(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0402
    name: Final = "Temperature Measurement"
    ep_attribute: Final = "temperature"

    class AttributeDefs(BaseAttributeDefs):
        # Temperature Measurement Information
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.int16s, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.int16s, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.int16s, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        # 0x0010: ('min_percent_change', UNKNOWN),
        # 0x0011: ('min_absolute_change', UNKNOWN),
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class PressureMeasurement(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0403
    name: Final = "Pressure Measurement"
    ep_attribute: Final = "pressure"

    class AttributeDefs(BaseAttributeDefs):
        # Pressure Measurement Information
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.int16s, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.int16s, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.int16s, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        # Extended attribute set
        scaled_value: Final = ZCLAttributeDef(id=0x0010, type=t.int16s, access="r")
        min_scaled_value: Final = ZCLAttributeDef(id=0x0011, type=t.int16s, access="r")
        max_scaled_value: Final = ZCLAttributeDef(id=0x0012, type=t.int16s, access="r")
        scaled_tolerance: Final = ZCLAttributeDef(
            id=0x0013, type=t.uint16_t, access="r"
        )
        scale: Final = ZCLAttributeDef(id=0x0014, type=t.int8s, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class FlowMeasurement(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0404
    name: Final = "Flow Measurement"
    ep_attribute: Final = "flow"

    class AttributeDefs(BaseAttributeDefs):
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class RelativeHumidity(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0405
    name: Final = "Relative Humidity Measurement"
    ep_attribute: Final = "humidity"

    class AttributeDefs(BaseAttributeDefs):
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class Occupancy(t.bitmap8):
    Unoccupied = 0b00000000
    Occupied = 0b00000001


class OccupancySensorType(t.enum8):
    PIR = 0x00
    Ultrasonic = 0x01
    PIR_and_Ultrasonic = 0x02
    Physical_Contact = 0x03


class OccupancySensorTypeBitmap(t.bitmap8):
    PIR = 0b00000001
    Ultrasonic = 0b00000010
    Physical_Contact = 0b00000100


class OccupancySensing(Cluster):
    Occupancy: Final = Occupancy
    OccupancySensorType: Final = OccupancySensorType
    OccupancySensorTypeBitmap: Final = OccupancySensorTypeBitmap

    cluster_id: Final[t.uint16_t] = 0x0406
    name: Final = "Occupancy Sensing"
    ep_attribute: Final = "occupancy"

    class AttributeDefs(BaseAttributeDefs):
        # Occupancy Sensor Information
        occupancy: Final = ZCLAttributeDef(
            id=0x0000, type=Occupancy, access="rp", mandatory=True
        )
        occupancy_sensor_type_bitmap: Final = ZCLAttributeDef(
            id=0x0001, type=t.bitmap8, access="r", mandatory=True
        )
        # PIR Configuration
        pir_o_to_u_delay: Final = ZCLAttributeDef(
            id=0x0010, type=t.uint16_t, access="rw"
        )
        pir_u_to_o_delay: Final = ZCLAttributeDef(
            id=0x0011, type=t.uint16_t, access="rw"
        )
        pir_u_to_o_threshold: Final = ZCLAttributeDef(
            id=0x0012, type=t.uint8_t, access="rw"
        )
        # Ultrasonic Configuration
        ultrasonic_o_to_u_delay: Final = ZCLAttributeDef(
            id=0x0020, type=t.uint16_t, access="rw"
        )
        ultrasonic_u_to_o_delay: Final = ZCLAttributeDef(
            id=0x0021, type=t.uint16_t, access="rw"
        )
        ultrasonic_u_to_o_threshold: Final = ZCLAttributeDef(
            id=0x0022, type=t.uint8_t, access="rw"
        )
        # Physical Contact Configuration
        physical_contact_o_to_u_delay: Final = ZCLAttributeDef(
            id=0x0030, type=t.uint16_t, access="rw"
        )
        physical_contact_u_to_o_delay: Final = ZCLAttributeDef(
            id=0x0031, type=t.uint16_t, access="rw"
        )
        physical_contact_u_to_o_threshold: Final = ZCLAttributeDef(
            id=0x0032, type=t.uint8_t, access="rw"
        )
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class LeafWetness(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0407
    name: Final = "Leaf Wetness Measurement"
    ep_attribute: Final = "leaf_wetness"

    class AttributeDefs(BaseAttributeDefs):
        # Leaf Wetness Measurement Information
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class SoilMoisture(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0408
    name: Final = "Soil Moisture Measurement"
    ep_attribute: Final = "soil_moisture"

    class AttributeDefs(BaseAttributeDefs):
        # Soil Moisture Measurement Information
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class PH(Cluster):
    cluster_id: Final[t.uint16_t] = 0x0409
    name: Final = "pH Measurement"
    ep_attribute: Final = "ph"

    class AttributeDefs(BaseAttributeDefs):
        # pH Measurement Information
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class ElectricalConductivity(Cluster):
    cluster_id: Final[t.uint16_t] = 0x040A
    name: Final = "Electrical Conductivity"
    ep_attribute: Final = "electrical_conductivity"

    class AttributeDefs(BaseAttributeDefs):
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class WindSpeed(Cluster):
    cluster_id: Final[t.uint16_t] = 0x040B
    name: Final = "Wind Speed Measurement"
    ep_attribute: Final = "wind_speed"

    class AttributeDefs(BaseAttributeDefs):
        # Wind Speed Measurement Information
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.uint16_t, access="rp", mandatory=True
        )
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.uint16_t, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.uint16_t, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.uint16_t, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class _ConcentrationMixin:
    """Mixin for the common attributes of the concentration measurement clusters"""

    class AttributeDefs(BaseAttributeDefs):
        measured_value: Final = ZCLAttributeDef(
            id=0x0000, type=t.Single, access="rp", mandatory=True
        )  # fraction of 1 (one)
        min_measured_value: Final = ZCLAttributeDef(
            id=0x0001, type=t.Single, access="r", mandatory=True
        )
        max_measured_value: Final = ZCLAttributeDef(
            id=0x0002, type=t.Single, access="r", mandatory=True
        )
        tolerance: Final = ZCLAttributeDef(id=0x0003, type=t.Single, access="r")
        cluster_revision: Final = foundation.ZCL_CLUSTER_REVISION_ATTR
        reporting_status: Final = foundation.ZCL_REPORTING_STATUS_ATTR


class CarbonMonoxideConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x040C
    name: Final = "Carbon Monoxide (CO) Concentration"
    ep_attribute: Final = "carbon_monoxide_concentration"


class CarbonDioxideConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x040D
    name: Final = "Carbon Dioxide (CO₂) Concentration"
    ep_attribute: Final = "carbon_dioxide_concentration"


class EthyleneConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x040E
    name: Final = "Ethylene (CH₂) Concentration"
    ep_attribute: Final = "ethylene_concentration"


class EthyleneOxideConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x040F
    name: Final = "Ethylene Oxide (C₂H₄O) Concentration"
    ep_attribute: Final = "ethylene_oxide_concentration"


class HydrogenConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0410
    name: Final = "Hydrogen (H) Concentration"
    ep_attribute: Final = "hydrogen_concentration"


class HydrogenSulfideConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0411
    name: Final = "Hydrogen Sulfide (H₂S) Concentration"
    ep_attribute: Final = "hydrogen_sulfide_concentration"


class NitricOxideConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0412
    name: Final = "Nitric Oxide (NO) Concentration"
    ep_attribute: Final = "nitric_oxide_concentration"


class NitrogenDioxideConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0413
    name: Final = "Nitrogen Dioxide (NO₂) Concentration"
    ep_attribute: Final = "nitrogen_dioxide_concentration"


class OxygenConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0414
    name: Final = "Oxygen (O₂) Concentration"
    ep_attribute: Final = "oxygen_concentration"


class OzoneConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0415
    name: Final = "Ozone (O₃) Concentration"
    ep_attribute: Final = "ozone_concentration"


class SulfurDioxideConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0416
    name: Final = "Sulfur Dioxide (SO₂) Concentration"
    ep_attribute: Final = "sulfur_dioxide_concentration"


class DissolvedOxygenConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0417
    name: Final = "Dissolved Oxygen (DO) Concentration"
    ep_attribute: Final = "dissolved_oxygen_concentration"


class BromateConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0418
    name: Final = "Bromate Concentration"
    ep_attribute: Final = "bromate_concentration"


class ChloraminesConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0419
    name: Final = "Chloramines Concentration"
    ep_attribute: Final = "chloramines_concentration"


class ChlorineConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x041A
    name: Final = "Chlorine Concentration"
    ep_attribute: Final = "chlorine_concentration"


class FecalColiformAndEColiFraction(_ConcentrationMixin, Cluster):
    """Percent of positive samples"""

    cluster_id: Final[t.uint16_t] = 0x041B
    name: Final = "Fecal coliform & E. Coli Fraction"
    ep_attribute: Final = "fecal_coliform_and_e_coli_fraction"


class FluorideConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = (
        0x041C  # XXX: spec repeats 0x041B but this seems like a mistake
    )
    name: Final = "Fluoride Concentration"
    ep_attribute: Final = "fluoride_concentration"


class HaloaceticAcidsConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x041D
    name: Final = "Haloacetic Acids Concentration"
    ep_attribute: Final = "haloacetic_acids_concentration"


class TotalTrihalomethanesConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x041E
    name: Final = "Total Trihalomethanes Concentration"
    ep_attribute: Final = "total_trihalomethanes_concentration"


class TotalColiformBacteriaFraction(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x041F
    name: Final = "Total Coliform Bacteria Fraction"
    ep_attribute: Final = "total_coliform_bacteria_fraction"


# XXX: is this a concentration? What are the units?
class Turbidity(_ConcentrationMixin, Cluster):
    """Cloudiness of particles in water where an average person would notice a 5 or higher"""

    cluster_id: Final[t.uint16_t] = 0x0420
    name: Final = "Turbidity"
    ep_attribute: Final = "turbidity"


class CopperConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0421
    name: Final = "Copper Concentration"
    ep_attribute: Final = "copper_concentration"


class LeadConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0422
    name: Final = "Lead Concentration"
    ep_attribute: Final = "lead_concentration"


class ManganeseConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0423
    name: Final = "Manganese Concentration"
    ep_attribute: Final = "manganese_concentration"


class SulfateConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0424
    name: Final = "Sulfate Concentration"
    ep_attribute: Final = "sulfate_concentration"


class BromodichloromethaneConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0425
    name: Final = "Bromodichloromethane Concentration"
    ep_attribute: Final = "bromodichloromethane_concentration"


class BromoformConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0426
    name: Final = "Bromoform Concentration"
    ep_attribute: Final = "bromoform_concentration"


class ChlorodibromomethaneConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0427
    name: Final = "Chlorodibromomethane Concentration"
    ep_attribute: Final = "chlorodibromomethane_concentration"


class ChloroformConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0428
    name: Final = "Chloroform Concentration"
    ep_attribute: Final = "chloroform_concentration"


class SodiumConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x0429
    name: Final = "Sodium Concentration"
    ep_attribute: Final = "sodium_concentration"


# XXX: is this a concentration? What are the units?
class PM25(_ConcentrationMixin, Cluster):
    """Particulate Matter 2.5 microns or less"""

    cluster_id: Final[t.uint16_t] = 0x042A
    name: Final = "PM2.5"
    ep_attribute: Final = "pm25"


class FormaldehydeConcentration(_ConcentrationMixin, Cluster):
    cluster_id: Final[t.uint16_t] = 0x042B
    name: Final = "Formaldehyde Concentration"
    ep_attribute: Final = "formaldehyde_concentration"
